/**********************************************************************\
 © COPYRIGHT 2013 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.
               
        The HANDLE.NET software is made available subject to the
         Handle System Public License, which may be obtained at
         http://hdl.handle.net/4263537/5030 or hdl:4263537/5030
\**********************************************************************/

package net.handle.hdllib;

import java.security.*;

public class SecretKeyAuthenticationInfo
  extends AuthenticationInfo
{
  private static Integer md5Lock = new Integer(123);
  private static MessageDigest md5 = null;

  private byte secretKey[];
  private byte userIdHandle[];
  private int userIdIndex;
  
  /** Construct a SecretKeyAuthenticationInfo object using the given user ID
    * handle with the given key at the given handle index.  This does not 
    * use a "shadow" of the given password */
  public SecretKeyAuthenticationInfo(byte userIDHandle[], int userIDIndex,
                                     byte secretKey[])
  {
    this.secretKey = secretKey;
    this.userIdHandle = userIDHandle;
    this.userIdIndex = userIDIndex;
  }
  
  
  /** Construct a SecretKeyAuthenticationInfo object using the given user ID
    * handle with the given key at the given handle index.  If the shadow 
    * parameter is true then the authentication will use a lower case 
    * hex encoded copy of the given secret key.  */
  public SecretKeyAuthenticationInfo(byte userIDHandle[], int userIDIndex,
                                     byte secretKey[], boolean shadow)
    throws Exception
  {
    this.secretKey = Encoder.encodeSecretKey(secretKey, shadow);
    this.userIdHandle = userIDHandle;
    this.userIdIndex = userIDIndex;
  }

  
  /***********************************************************************
   * Get the identifier for the type of authentication performed.  In this
   * case, the authentication type is AT_SECRET_KEY.
   ***********************************************************************/
  public byte[] getAuthType()
  {
    return Common.MD5_SECRET_KEY_TYPE;
  }


  /***********************************************************************
   * Sign the given nonce and requestDigest given as a challenge to the
   * given request.  The implementation of this method should also probably
   * verify that the client did in fact send the specified request, and 
   * that the associated digest is a valid digest of the request.
   * @return a digest of the concatenation of secret key, nonce,
   * requestDigest, and secret key.
   ***********************************************************************/
  public byte[] authenticate(ChallengeResponse challenge, AbstractRequest request)
    throws HandleException
  {
    // need to verify that this is actually a digest of the specified request
    byte origMessage[] = request.getEncodedMessageBody();
    byte origDigest[] = Util.doDigest(challenge.rdHashType, origMessage);

    if(!Util.equals(origDigest, challenge.requestDigest)) {
      throw new HandleException(HandleException.SECURITY_ALERT,
                                "Asked to sign unidentified request!");
    }

    // create a hash of the secret key, nonce, and request digest
    byte digestAlg = Common.DEFAULT_HASH_CODE;
    boolean oldFormat = !challenge.hasEqualOrGreaterVersion(2,1);
    if (oldFormat) {
      digestAlg = Common.HASH_CODE_MD5_OLD_FORMAT;
    }
    
    byte digest[] = Util.doDigest(digestAlg, secretKey,
                                  challenge.nonce, challenge.requestDigest,
                                  secretKey);
    if(oldFormat) {
      return digest;
    } else {
      byte authResponse[] = new byte[digest.length + 1];
      authResponse[0] = digestAlg;
      System.arraycopy(digest, 0, authResponse, 1, digest.length);
      return authResponse;
    }
  }


  /***********************************************************************
   * Get the name of the handle that identifies the user that is 
   * represented by this authentication object.
   ***********************************************************************/
  public byte[] getUserIdHandle() {
    return userIdHandle;
  }

  /***********************************************************************
   * Get the index of the handle value that identifies this user.
   * The returned index value of the handle that identifies this user
   * should contain a value with a type (public key, secret key, etc)
   * that corresponds to the way that this user is authenticating.
   ***********************************************************************/
  public int getUserIdIndex()
  {
    return userIdIndex;
  }
  

  /** Return the byte-encoded representation of the secret key. */
  public byte[] getSecretKey() {
    return secretKey;
  }

  public String toString() {
    return "secret_key:"+String.valueOf(userIdIndex)+':'+
      ((userIdHandle==null)?"null":Util.decodeString(userIdHandle));
  }

  public boolean equals(Object obj) {
      if(obj==null) return false;
      if(!(obj instanceof SecretKeyAuthenticationInfo)) return false;
      if(super.equals(obj)) return true;
      SecretKeyAuthenticationInfo other = (SecretKeyAuthenticationInfo) obj;
      boolean sameIdentity = Util.equals(getUserIdHandle(),other.getUserIdHandle()) && getUserIdIndex()==other.getUserIdIndex();
      if(!sameIdentity) return false;
      return Util.equals(secretKey,other.secretKey);
  }
}
