/**********************************************************************\
 © COPYRIGHT 2013 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.
               
        The HANDLE.NET software is made available subject to the
         Handle System Public License, which may be obtained at
         http://hdl.handle.net/4263537/5030 or hdl:4263537/5030
\**********************************************************************/

package net.handle.hdllib;

import java.security.*;
import java.io.*;

/**************************************************************************
 * Objects of this class can be used to verify the contents of a stream
 * generated by a SignedOutputStream class.
 *
 * Note: This is not a part of the official handle system specification.
 * This class and the format of the stream it reads will almost definitely
 * change with future versions of the handle system.  A possible improvement
 * is to use digest or MAC-based verification rather than full public/private
 * key signatures for each block.
 **************************************************************************/
public class SignedInputStream
  extends InputStream
{
  public static final int STREAM_TYPE_PK = 0;

  private InputStream in;
  private Signature sig;
  private int streamType;

  /**************************************************************************
   * Create a stream that can verify the data read from the stream in
   * blocks.  The caller should call the verifyBlock method at the end
   * of every block of data that needs to be verified.  verifyBlock must 
   * be called at the same position in the stream that signBlock was called
   * in the parallel SignedOutputStream object that generated the stream.
   **************************************************************************/
  public SignedInputStream(PublicKey sourceKey, InputStream in)
    throws Exception
  {
    super();
    this.in = in;
    
    // read the type-of-stream identifier - the first 4 bytes
    byte streamTypeBuf[] = new byte[Encoder.INT_SIZE];
    if(in.read(streamTypeBuf) < streamTypeBuf.length)
      throw new SignatureException("End of stream while reading stream type");
    streamType = Encoder.readInt(streamTypeBuf,0);

    // initialize ourself depending on the type of stream
    switch(streamType) {
      case STREAM_TYPE_PK:
        // read the hash ID so we know how to verify the signatures...
        byte lenbuf[] = new byte[Encoder.INT_SIZE];

        if(in.read(lenbuf)<lenbuf.length)
          throw new SignatureException("Unable to read signature hash type");

        int hashIDLen = Encoder.readInt(lenbuf,0);
        if(hashIDLen < 0 || hashIDLen > Common.MAX_ARRAY_SIZE)
          throw new SignatureException("Invalid hash ID - too long");

        byte hashID[] = new byte[hashIDLen];
        if(in.read(hashID) < hashIDLen)
          throw new SignatureException("End of stream while reading hash ID");

        sig = 
          Signature.getInstance(Util.getSigIdFromHashAlgId(hashID, sourceKey.getAlgorithm()));
        sig.initVerify(sourceKey);

        break;
      default:
        throw new SignatureException("Unrecognized stream type: "+streamType);
    }
    
  }

  /**************************************************************************
   * Read a byte from the stream.
   **************************************************************************/
  public int read()
    throws IOException
  {
    int b = in.read();
    if(b<0) return b;
    try {
      sig.update((byte)b);
    } catch (SignatureException e) {
      throw new IOException("Error updating signature: "+e);
    }
    return b;
  }


  /**************************************************************************
   * Reads a signature from the stream and verifies the bytes read since
   * the last verification based on that signature.  This should be called
   * at the exact same point in the stream as the signBlock() in the
   * SignedOutpuStream class.
   **************************************************************************/
  public boolean verifyBlock()
    throws IOException,
           SignatureException
  {
    byte lenbuf[] = new byte[Encoder.INT_SIZE];
    int r;
    int n = 0;
    while(n<lenbuf.length && (r=in.read(lenbuf, n, lenbuf.length-n))>=0)
      n += r;
    if(n < Encoder.INT_SIZE)
      throw new SignatureException("End of stream while reading signature length");
    
    int sigLen = Encoder.readInt(lenbuf,0);
    if(sigLen < 0 || sigLen > Common.MAX_ARRAY_SIZE)
      throw new SignatureException("Invalid signature - too long");

    byte buf[] = new byte[sigLen];
    n = 0;
    while(n<sigLen && (r=in.read(buf, n, sigLen-n))>=0)
      n += r;

    if (n < sigLen)
      throw new SignatureException("End of stream while reading signature."+
                                   "  Expected "+sigLen+",  got "+n);
    
    return sig.verify(buf);
  }

}

