/**********************************************************************\
 © COPYRIGHT 2013 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.
               
        The HANDLE.NET software is made available subject to the
         Handle System Public License, which may be obtained at
         http://hdl.handle.net/4263537/5030 or hdl:4263537/5030
\**********************************************************************/

package net.handle.server;

import net.cnri.io.AtomicFile;
import net.cnri.util.StreamTable;
import net.handle.security.*;
import net.handle.server.replication.ReplicationDaemon;
import net.handle.server.MonitorDaemon;
import net.handle.util.IntTable;
import net.handle.apps.simple.SiteInfoConverter;
import net.handle.hdllib.*;

import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.io.*;
import java.security.*;
import java.security.interfaces.RSAPublicKey;

import javax.crypto.interfaces.*;
import javax.crypto.spec.*;
import java.util.Properties;


public class HandleServer
extends AbstractServer
{
    private static final boolean DEBUG_AUTHENTICATION = false;
    
	private static final byte MSG_INTERNAL_ERROR[] =
		Util.encodeString("Internal Error");
	private static final byte MSG_NOT_A_PRIMARY[] =
		Util.encodeString("Not a primary server");
	private static final byte MSG_SERVER_TEMPORARILY_DISABLED[] =
		Util.encodeString("Server temporarily disabled");
	private static final byte MSG_WRONG_SERVER_HASH[] =
		Util.encodeString("Request was hashed incorrectly");
	private static final byte MSG_NA_NOT_HOMED_HERE[] =
		Util.encodeString("That naming authority doesn't live here");
	private static final byte MSG_INDEXES_MUST_BE_POSITIVE[] =
		Util.encodeString("Value indexes must be non-negative");
	private static final byte MSG_EMPTY_VALUE_LIST[] =
		Util.encodeString("Value list was empty");
	private static final byte MSG_READ_ONLY_VALUE[] =
		Util.encodeString("Value is read-only");
	private static final byte MSG_NOT_A_NA_HANDLE[] =
		Util.encodeString("Handle is not a naming authority handle");
	private static final byte MSG_INVALID_ENCODING[] =
		Util.encodeString("Invalid UTF8 encoding");
	private static final byte MSG_INVALID_NA_HANDLE[] =
		Util.encodeString("Invalid naming authority handle");
	private static final byte MSG_SERVER_BACKUP[] =
		Util.encodeString("Server can only resolve handles due to maintenance. Come back later");

	private static final byte MSG_NEED_RSA_EXCHANGEKEY[] =
		Util.encodeString("Invalid exchange key type. Need RSA public key.");
	private static final byte MSG_INVALID_SESSION_OR_TIMEOUT[] =
		Util.encodeString("Invalid session id or session time out. Please try again.");
	private static final byte MSG_NEED_LIST_HDLS_PERM[] =
		Util.encodeString("This server does not support the list handles operation.");
	private static final byte SERVER_STATUS_HANDLE[] =
		Util.encodeString("0.SITE/status");
	private static final byte SERVER_STATUS_HDL_TYPE[] =
		Util.encodeString("CNRI.SERVER_STATUS");
	private static final byte MSG_SESSION_REQUIRED[] =
		Util.encodeString("Sessions are required for administration on this server");

	public static final String CASE_SENSITIVE = "case_sensitive";
	public static final String ENABLE_STATUS_HDL = "enable_status_handle";
    public static final String ENABLE_MONITOR_DAEMON = "enable_monitor_daemon";
    public static final String ENABLE_HOMED_PREFIX_LOCAL_SITES = "enable_homed_prefix_na_lookup_optimization";
	public static final String SERVER_ADMIN_FULL_ACCESS = "server_admin_full_access";
	public static final String ANONYMOUS_FULL_ACCESS = "anonymous_admin_full_access";
	public static final String MAX_AUTH_TIME = "max_auth_time";
	public static final String THIS_SERVER_ID = "this_server_id";
	public static final String IS_PRIMARY = "is_primary";
	public static final String SERVER_ADMINS = "server_admins";
    public static final String HOME_ONLY_ADMINS = "home_only_admins";
	public static final String BACKUP_ADMINS = "backup_admins";
	public static final String REPLICATION_ADMINS = "replication_admins";
	public static final String DO_RECURSION = "allow_recursion";
	public static final String ALLOW_NA_ADMINS = "allow_na_admins";
	public static final String READ_ONLY_TXN_QUEUE = "read_only_txn_queue";
	public static final String ALLOW_LIST_HANDLES = "allow_list_hdls";
	public static final String PREFERRED_GLOBAL = "preferred_global";
	public static final String MAX_SESSION_TIME = "max_session_time";
	public static final String REQUIRE_SESSIONS = "require_sessions";
	public static final String ENCRYPTION_ALGORITHM = "encryption_alg";

	public static final String TEMPLATE_DELIM_CONFIG_KEY = "template_delimiter";
	public static final String NAMESPACE_CONFIG_KEY = "namespace";

	public static final String DB_TXN_QUEUE_DIR = "dbtxns";
	public static final String TXN_QUEUE_DIR = "txns";
	public static final String TXN_ID_FILE = "txn_id";
	public static final String STORAGE_FILE = "handles.jdb";
	public static final String NA_STORAGE_FILE = "nas.jdb";
	public static final String CACHE_STORAGE_FILE = "cache.jdb";
	public static final String STORAGE_FILE_BACKUP = "handles.jdb.backup";
	public static final String NA_STORAGE_FILE_BACKUP = "nas.jdb.backup";
	public static final String SITE_INFO_FILE = "siteinfo.bin";
    public static final String SITE_INFO_JSON_FILE = "siteinfo.json";
    public static final String PUBLIC_KEY_FILE = "pubkey.bin";
    public static final String PRIVATE_KEY_FILE = "privkey.bin";
	public static final String DO_REPLICATION = "do_replication";

	public static final int RECURSION_LIMIT = 10;
	public static final int LIST_HANDLES_PER_MSG = 50;

	public static final String DEFAULT_ENC_ALG = "DES"; // options: DES, AES, DESEDE

	private String TRANSACTION_LOCK = "TRANSACTION_LOCK";
	private String NEXT_TXN_ID_LOCK = "NEXT_TXN_ID_LOCK";

	private static final int NUM_SERVER_SIGNATURES = 50;

	private static final int DEL_HANDLE_PERM[] = { AdminRecord.DELETE_HANDLE };
	private static final int ADD_HANDLE_PERM[] = { AdminRecord.ADD_HANDLE };
	private static final int READ_VAL_PERM[] = { AdminRecord.READ_VALUE };
	private static final int ADD_ADM_PERM[] = { AdminRecord.ADD_ADMIN };
	private static final int ADD_VAL_PERM[] = { AdminRecord.ADD_VALUE };
	private static final int ADD_ADM_AND_VAL_PERM[] = { AdminRecord.ADD_ADMIN,
		AdminRecord.ADD_VALUE };
	private static final int REM_VAL_PERM[] = { AdminRecord.REMOVE_VALUE };
	private static final int REM_ADM_PERM[] = { AdminRecord.REMOVE_ADMIN };
	private static final int REM_ADM_AND_VAL_PERM[] = { AdminRecord.REMOVE_ADMIN,
		AdminRecord.REMOVE_VALUE };
	private static final int MOD_ADM_PERM[] = { AdminRecord.MODIFY_ADMIN };
	private static final int MOD_VAL_PERM[] = { AdminRecord.MODIFY_VALUE };
	private static final int ADM_TO_VAL_PERM[] = { AdminRecord.REMOVE_ADMIN,
		AdminRecord.MODIFY_VALUE };
	private static final int VAL_TO_ADM_PERM[] = { AdminRecord.MODIFY_VALUE,
		AdminRecord.ADD_ADMIN };

	private static final int ADD_SUB_NA_PERM[] = { AdminRecord.ADD_NAMING_AUTH };
	private static final int LIST_HDLS_PERM[] = { AdminRecord.LIST_HANDLES };

	private static final byte SIGN_TEST[] = Util.encodeString("Testing...1..2..3");

	private GenericRequest nextTxnIdRequest;

	//private volatile boolean keepRunning = true; // moved to AbstractServer
	private volatile boolean serverEnabled = true;
	private static int nextAuthId = 0;
	private long maxAuthTime;
	private HandleStorage storage;
	private IntTable pendingAuthorizations;
	private boolean caseSensitive = false;
	private boolean serverAdminFullAccess = false;
	private boolean wideOpenAccessAllowed = false;
	private ValueReference serverAdmins[] = {};
        private ValueReference homeOnlyAdmins[] = {};
	private ValueReference backupAdmins[] = {};
	private ValueReference replicationAdmins[] = {};
	private boolean requireSessions = false;

	// legacy
	private String configTemplateDelimiter = null; // overridden in constructor
	private NamespaceInfo configNamespaceInfo = null;

	private TransactionQueueInterface txnQueue;
	private int currentSigIndex = 0;
	private SiteInfo thisSite = null;
	private int thisServerNum = -1;

	private PublicKey publicKey = null;
	private PrivateKey privateKey = null;
	private Signature serverSignatures[] = null;
	private boolean allowRecursiveQueries = false;
	private boolean allowNAAdmins = true;
	private boolean allowListHdls = true;
	private String preferredGlobal = null;
	private boolean isPrimary = false;
	private boolean multiPrimary = false;
	private boolean doReplication = false;
	private long nextTxnId = 0;
	private File txnIdFile = null;
	private boolean nextTxnIdInitialized = false;

	private ReplicationDaemon replicationDaemon;
	private long startTime = 0;
	private AtomicLong numRequests = new AtomicLong();
        private AtomicLong numResolutionRequests = new AtomicLong();
        private AtomicLong numAdminRequests = new AtomicLong();
        private AtomicLong numTxnRequests = new AtomicLong();
	private boolean enableStatusHandle = true;
	private boolean enableHomedPrefixNaLookupOptimization = true;
	private int encryptionAlgorithm;
	private Object lockHash[] = { new Object() };

	// a server side session manager to manage session
	private final SessionManager sessions = new SessionManager();
	private String writeLock = null;

	private MonitorDaemon monitorDaemon;

	private PublicKey getPublicKeyFromFile() {
		File publicKeyFile = new File(main.getConfigDir(), PUBLIC_KEY_FILE);
		PublicKey result = null;
		try {
			result = Util.getPublicKeyFromFile(publicKeyFile.getAbsolutePath());
		} catch (Exception e) {
			e.printStackTrace();
		}
		return result;
	}
	
	public PublicKey getPublicKey() {
	    return publicKey;
	}
	
	public PrivateKey getPrivateKey() {
		return privateKey;
	}

	protected HandleServer() {
	    // for testing
	}
	
	public HandleServer(Main main, StreamTable config)
	throws Exception
	{
		super(main, config);
		this.pendingAuthorizations = new IntTable();
		this.startTime = System.currentTimeMillis();
		this.enableStatusHandle = config.getBoolean(ENABLE_STATUS_HDL, false);
		this.enableHomedPrefixNaLookupOptimization = config.getBoolean(ENABLE_HOMED_PREFIX_LOCAL_SITES,true);
        
		if(config.getBoolean(ENABLE_MONITOR_DAEMON,false)) {
		    monitorDaemon = new MonitorDaemon(60,
		            startTime,
		            numRequests,
		            numResolutionRequests,
		            numAdminRequests,
		            numTxnRequests);
		    monitorDaemon.start();
		}

		writeLock = String.valueOf((new Random()).nextInt());
		requireSessions = config.getBoolean(REQUIRE_SESSIONS, false);
		String encAlg = config.getStr(ENCRYPTION_ALGORITHM, DEFAULT_ENC_ALG).toLowerCase().trim();
		if(encAlg.equals("des")) {
			encryptionAlgorithm = HdlSecurityProvider.ENCRYPT_ALG_DES;
		} else if(encAlg.equals("desede")) {
			encryptionAlgorithm = HdlSecurityProvider.ENCRYPT_ALG_DESEDE;
		} else if(encAlg.equals("aes")) {
			encryptionAlgorithm = HdlSecurityProvider.ENCRYPT_ALG_AES;
		} else {
			throw new Exception("Invalid encryption algorithm: '"+encAlg+
			"'; Please use either des, desede, or aes");
		}

		if(config.containsKey(DO_RECURSION))
			allowRecursiveQueries = config.getBoolean(DO_RECURSION);

		if(config.containsKey(SERVER_ADMIN_FULL_ACCESS))
			serverAdminFullAccess = config.getBoolean(SERVER_ADMIN_FULL_ACCESS);
		if(config.containsKey(ANONYMOUS_FULL_ACCESS))
		  wideOpenAccessAllowed = config.getBoolean(ANONYMOUS_FULL_ACCESS);
		if(config.containsKey(ALLOW_LIST_HANDLES))
			allowListHdls = config.getBoolean(ALLOW_LIST_HANDLES);

		if(config.containsKey(ALLOW_NA_ADMINS))
			allowNAAdmins = config.getBoolean(ALLOW_NA_ADMINS);

		if(config.containsKey(PREFERRED_GLOBAL)){
			preferredGlobal = config.getStr(PREFERRED_GLOBAL);
			Properties p = new Properties(System.getProperties());
			p.put("hdllib.preferredGlobal",preferredGlobal);
			System.setProperties(p);
		}

		// get the site info that describes this site that this server is
		// a part of
		try {
			// get the index of this server in the site info
			int thisId = Integer.parseInt((String)config.get(THIS_SERVER_ID));

			// get the site information from the site-info file
			SiteInfo site;
			File siteInfoFile = new File(main.getConfigDir(), SITE_INFO_FILE);
			if(!siteInfoFile.exists() || !siteInfoFile.canRead()) {
			    File siteInfoJsonFile = new File(main.getConfigDir(), SITE_INFO_JSON_FILE);
			    if(!siteInfoJsonFile.exists() || !siteInfoJsonFile.canRead()) {
			        System.err.println("Missing or inaccessible site info file: "+
			                siteInfoFile.getAbsolutePath() + "/" + siteInfoJsonFile.getName());
			        throw new Exception("Missing or inaccessible site info file: "+
			                siteInfoFile.getAbsolutePath() + "/" + siteInfoJsonFile.getName());
			    } else {
			        byte[] siteInfoBytes = Util.getBytesFromFile(siteInfoJsonFile);
			        try {
			            site = SiteInfoConverter.convertToSiteInfo(new String(siteInfoBytes,"UTF-8"));
			        } catch(Throwable t) {
	                    System.err.println("Missing or inaccessible site info file: "+
	                            siteInfoFile.getAbsolutePath() + "/" + siteInfoJsonFile.getName());
	                    throw new Exception("Missing or inaccessible site info file: "+
	                            siteInfoFile.getAbsolutePath() + "/" + siteInfoJsonFile.getName(),t);
			        }
			    }
			}
			else {
			    site = new SiteInfo();
			    byte siteInfoBuf[] = new byte[(int)siteInfoFile.length()];
			    new FileInputStream(siteInfoFile).read(siteInfoBuf);
			    Encoder.decodeSiteInfoRecord(siteInfoBuf, 0, site);
			}
			
			thisServerNum = -1;
			for(int i=0; i<site.servers.length; i++) {
				if(site.servers[i].serverId==thisId) {
					thisServerNum = i;
				}
			}

			if(thisServerNum<0) {
				throw new Exception("Server ID "+thisId+" does not exist in site!");
			}
			thisSite = site;
		} catch (Exception e) {
			System.err.println("Invalid site/server specification: "+e);
			throw e;
		}

		// this secret key bytes will be used to encrypt/decrypt both server key
		// and session exchange key
		byte secKey[] = null;
		try {
		    publicKey = getPublicKeyFromFile();
		    
		    // read the private key from the private key file...
			File privateKeyFile = new File(main.getConfigDir(), PRIVATE_KEY_FILE);
			if(!privateKeyFile.exists() || !privateKeyFile.canRead()) {
				System.err.println("Missing or inaccessible private key file: "+
						privateKeyFile.getAbsolutePath());
				throw new Exception("Missing or inaccessible private key file: "+
						privateKeyFile.getAbsolutePath());
			}
			FileInputStream in = new FileInputStream(privateKeyFile);
			byte encKeyBytes[] = new byte[(int)privateKeyFile.length()];
			int n=0;
			int r;
			while(n<encKeyBytes.length && (r=in.read(encKeyBytes,n,encKeyBytes.length-n))>=0) {
				n += r;
			}

			byte keyBytes[] = null;
			if(Util.requiresSecretKey(encKeyBytes)) {
				// ask for a secret key to decrypt the private key get the passphrase
				// and decrypt the server's private key, it will also be used for
				// encrpting/decrypt the RSA/DH session exchange keys
				secKey = Util.getPassphrase("Enter the passphrase for this server's authentication private key: ");
			}

			keyBytes = Util.decrypt(encKeyBytes, secKey);

			try {
				privateKey = Util.getPrivateKeyFromBytes(keyBytes,0);
			}
			catch (Exception e) {
				System.err.println("\n**********************************************"+
						"****************************"+
						"\nError parsing private key, please make sure "+
						"the passphrase is correct.\n"+
						"***************************"+
				"***********************************************\n");
				throw e;
			}

			//clear the private key bytes in memory
			for(int i=0; i<keyBytes.length; i++) keyBytes[i] = (byte)0;

			// create a bunch of the signature objects so that signing responses
			// doesn't become a bottleneck
			serverSignatures = new Signature[NUM_SERVER_SIGNATURES];
			for(int i=0; i<serverSignatures.length; i++) {
				serverSignatures[i] = Signature.getInstance(Util.getSigIdFromHashAlgId(Common.HASH_ALG_SHA1, privateKey.getAlgorithm()));
				serverSignatures[i].initSign(privateKey);
			}

			// the signature has been initialized... now let's verify that it matches
			// the signature in the site information for this server.
			PublicKey pubKey = thisSite.servers[thisServerNum].getPublicKey();

			// generate a test signature,
			serverSignatures[0].update(SIGN_TEST);
			byte testSig[] = serverSignatures[0].sign();

			// verify the test signature
			Signature verifier = Signature.getInstance(serverSignatures[0].getAlgorithm());
			verifier.initVerify(pubKey);

			verifier.update(SIGN_TEST);
			if(!verifier.verify(testSig)) {
				throw new Exception("Private key doesn't match public key from site info!");
			}

		} catch (Exception e) {
			System.err.println("Unable to initialize server signature object: "+e);
			e.printStackTrace(System.err);
			throw e;
		}

		SessionManager.initializeSessionKeyRandom();

		sessions.checkTimeoutSession();

		int maxSessionTimeout = Common.DEFAULT_SESSION_TIMEOUT;
		if (config.containsKey(MAX_SESSION_TIME)){
			try {
				maxSessionTimeout = Integer.parseInt(String.valueOf(config.get(MAX_SESSION_TIME)).trim());
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
				"Invalid session timeout allowance.  Using default (24 hours)");
			}
		}
		if (maxSessionTimeout < 60) { //less than one minute, set it to 1 minute
			maxSessionTimeout = 60;
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
			"Adjusted session timeout allowance. Using 1 minute.");
		}
		SessionInfo.setDefaultTimeout(maxSessionTimeout);

		// clear the secret key
		for(int i=0; secKey!=null && i<secKey.length; i++) secKey[i] = (byte)0;

		isPrimary = thisSite.isPrimary;
		multiPrimary = thisSite.multiPrimary;

		// Get the next transaction ID, and prepare the file for updating
		txnIdFile = new File(main.getConfigDir(), TXN_ID_FILE);
		AtomicFile txnIdAtomicFile = new AtomicFile(txnIdFile);
		Reader fin = null;
		try {
		    if(txnIdFile.exists()) {
				char ch[] = new char[512];
				fin = new InputStreamReader(txnIdAtomicFile.openRead(),"UTF-8");
				int r, n=0;
				while((r=fin.read(ch, n, ch.length-n))>=0)
					n+= r;
				nextTxnId = Long.parseLong(new String(ch,0,n).trim());
				nextTxnIdInitialized = true;
			} else {
				nextTxnId = 0; // it will be incremented to 1 before use...
				nextTxnIdInitialized = false;
			}
		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL, "Invalid transaction ID! "+e);
			throw e;
		} finally {
			if(fin!=null) {
				try { fin.close(); } catch (Exception e){}
			}
		}

		if(config.containsKey(SERVER_ADMINS)) {
			try {
				Vector adminVect = (Vector)config.get(SERVER_ADMINS);
				serverAdmins = new ValueReference[adminVect.size()];
				for(int i=0; i<adminVect.size(); i++) {
					String adminStr = String.valueOf(adminVect.elementAt(i));
					int colIdx = adminStr.indexOf(':');
					if(colIdx<=0)
						throw new Exception("Invalid server administrator ID: \""+
								adminStr+"\"");
					serverAdmins[i] =
						new ValueReference(Util.encodeString(adminStr.substring(colIdx+1)),
								Integer.parseInt(adminStr.substring(0,colIdx)));
				}
			} catch (Exception e) {
				throw new Exception("Error processing server administrator list: "+e);
			}
		}

        if(config.containsKey(HOME_ONLY_ADMINS)) {
            try {
                Vector adminVect = (Vector)config.get(HOME_ONLY_ADMINS);
                homeOnlyAdmins = new ValueReference[adminVect.size()];
                for(int i=0; i<adminVect.size(); i++) {
                    String adminStr = String.valueOf(adminVect.elementAt(i));
                    int colIdx = adminStr.indexOf(':');
                    if(colIdx<=0)
                        throw new Exception("Invalid home-only administrator ID: \""+
                                adminStr+"\"");
                    homeOnlyAdmins[i] =
                        new ValueReference(Util.encodeString(adminStr.substring(colIdx+1)),
                                Integer.parseInt(adminStr.substring(0,colIdx)));
                }
            } catch (Exception e) {
                throw new Exception("Error processing home-only administrator list: "+e);
            }
        }

		if(config.containsKey(BACKUP_ADMINS)) {
			try {
				Vector adminVect = (Vector)config.get(BACKUP_ADMINS);
				backupAdmins = new ValueReference[adminVect.size()];
				for(int i=0; i<adminVect.size(); i++) {
					String adminStr = String.valueOf(adminVect.elementAt(i));
					int colIdx = adminStr.indexOf(':');
					if(colIdx<=0)
						throw new Exception("Invalid server backup administrator ID: \""+
								adminStr+"\"");
					backupAdmins[i] =
						new ValueReference(Util.encodeString(adminStr.substring(colIdx+1)),
								Integer.parseInt(adminStr.substring(0,colIdx)));
				}
			} catch (Exception e) {
				throw new Exception("Error processing server backup administrator list: "+e);
			}
		}


		if(config.containsKey(REPLICATION_ADMINS)) {
			try {
				Vector adminVect = (Vector)config.get(REPLICATION_ADMINS);
				replicationAdmins = new ValueReference[adminVect.size()];
				for(int i=0; i<adminVect.size(); i++) {
					String adminStr = String.valueOf(adminVect.elementAt(i));
					int colIdx = adminStr.indexOf(':');
					if(colIdx<=0)
						throw new Exception("Invalid replication administrator ID: \""+
								adminStr+"\"");
					replicationAdmins[i] =
						new ValueReference(Util.encodeString(adminStr.substring(colIdx+1)),
								Integer.parseInt(adminStr.substring(0,colIdx)));
				}
			} catch (Exception e) {
				throw new Exception("Error processing replication administrator list: "+e);
			}
		} else {
			replicationAdmins = new ValueReference[0];
		}

		doReplication = config.getBoolean(DO_REPLICATION, multiPrimary || !isPrimary);

		if(doReplication) {
		  replicationDaemon = new ReplicationDaemon(main,this,config,main.getConfigDir());
		}

		if(isPrimary) {
			// create the transaction queue.
			if(config.getBoolean(READ_ONLY_TXN_QUEUE, false)) {
				txnQueue = new ReadOnlyTransactionQueue(new File(main.getConfigDir(), TXN_QUEUE_DIR));
			} else {
				txnQueue = new TransactionQueue(new File(main.getConfigDir(), TXN_QUEUE_DIR));
			}
		}

		caseSensitive = config.getBoolean(CASE_SENSITIVE);

		configTemplateDelimiter = config.getStr(TEMPLATE_DELIM_CONFIG_KEY, "").trim();
		String configNamespaceInfoStr = config.getStr(NAMESPACE_CONFIG_KEY, null);
		if(configNamespaceInfoStr!=null) {
			configNamespaceInfo = new NamespaceInfo(configNamespaceInfoStr.getBytes("UTF-8"));
		}

		// load the database/storage system
		storage = HandleStorageFactory.getStorage(main.getConfigDir(), config, isPrimary);

		try {
			maxAuthTime = Long.parseLong(String.valueOf(config.get(MAX_AUTH_TIME)).trim());
		} catch (Exception e) {
			System.err.println("Invalid authentication time allowance.  "+
			"Using default (20 seconds)");
			maxAuthTime = 20000;
		}

		// add local site info to resolver configuration to avoid trips to global
		// when we are only talking to our peers
		if(enableHomedPrefixNaLookupOptimization) {
		    final SiteInfo ss[] = { thisSite };
		    storage.scanNAs(new ScanCallback() {
		        public void scanHandle(byte handle[]) {
		            resolver.getConfiguration()
		            .setLocalSites(Util.decodeString(handle), ss);
		        }
		    });
		}

		// expand the lockHash so that more write operations can happen in parallel
		lockHash = new Object[256];
		for(int i=0; i<lockHash.length; i++) {
			lockHash[i] = new Object();
		}

		ChallengeResponse.initializeRandom();

		nextTxnIdRequest = new GenericRequest(Common.SERVER_TXN_ID_HANDLE,
				AbstractMessage.OC_GET_NEXT_TXN_ID,
				null);
		nextTxnIdRequest.certify = true;

		// start thread to purge timed-out pendingAuthorizations.
		ChallengePurgeThread cpt = new ChallengePurgeThread();
		cpt.setDaemon(true);
		cpt.setPriority(Thread.MIN_PRIORITY);
		cpt.start();

		if(doReplication) {
		  System.err.println("starting replication thread");
			replicationDaemon.setDaemon(true);
			replicationDaemon.setPriority(Thread.MIN_PRIORITY);
			replicationDaemon.start();
		}
	}


  /**
   * Tell the server to re-dump all handles from a primary handle server
   */
	public void dumpHandles()
	  throws HandleException, IOException
	{
	  replicationDaemon.dumpHandles(true);
  }




	/** Return the siteinfo for this server. */
	public SiteInfo getSiteInfo() {
		return thisSite;
	}

	public int getServerNum() {
	    return thisServerNum;
	}

	/** Return the server information for this server */
	public ServerInfo getServerInfo() {
		return thisSite.servers[thisServerNum];
	}

	public HandleStorage getStorage() {
	    return this.storage;
	}

	/** Get handle values, possibly modified by the template system */
	public final byte[][] getRawHandleValuesWithTemplate(byte inHandle[], int indexList[],
			byte typeList[][], short recursionCount)
	throws HandleException
	{
		byte[][] values = storage.getRawHandleValues(caseSensitive ? inHandle : Util.upperCase(inHandle),indexList,typeList);
		if(values!=null) return values;

		// if there is no overridden handle value, return a calculated value
		// config value overrides
		String templateDelimiter = null;
		NamespaceInfo ns = null;
        ResolutionRequest resReq = new ResolutionRequest(inHandle,null,null,null);
		// request used for service information lookup
		resReq.certify = true;
		resReq.recursionCount = recursionCount;
		// use namespace from prefix, then namespace from config.dct, then template_delimiter from config.dct
		// TODO this interacts oddly with independent handle service; must put prefix in the independent server
		ns = resolver.getNamespaceInfo(resReq);
		if(ns==null) ns = configNamespaceInfo;
		if(ns!=null) templateDelimiter = ns.templateDelimiter();
		if(templateDelimiter==null || templateDelimiter.length()==0) templateDelimiter = configTemplateDelimiter;
		if(templateDelimiter==null || templateDelimiter.length()==0) return null;
		
		byte[] naHandle = resolver.getNAHandle(resReq);
        if(naHandle==null) {
		    int i;
		    for (i = 0; i<inHandle.length; i++) {
		        if(inHandle[i]=='/') break;
		    }
		    if(i>=inHandle.length) return null;
		    naHandle = new byte[i];
		    System.arraycopy(inHandle,0,naHandle,0,i);
		}
		String naHandleStr;
		// find the first delimiter after the NA portion of the handle.
		if(Util.startsWith(naHandle,Common.NA_HANDLE_PREFIX)) naHandleStr = Util.decodeString(naHandle,Common.NA_HANDLE_PREFIX.length,naHandle.length - Common.NA_HANDLE_PREFIX.length);
		else naHandleStr = Util.decodeString(naHandle);
		String inHandleStr = Util.decodeString(inHandle);
		int startDelim = inHandleStr.indexOf(templateDelimiter,naHandleStr.length());
		if(startDelim<0) return null;

		boolean allValues = (indexList==null || indexList.length==0) && (typeList==null || typeList.length==0);

		String baseHandleStr = inHandleStr.substring(0,startDelim);
		String afterDelimStr = inHandleStr.substring(startDelim + templateDelimiter.length());
		byte[] baseHandle = Util.encodeString(baseHandleStr);
		if( thisSite.servers.length > 1 &&
				thisSite.determineServerNum(baseHandle)!=thisServerNum) {
		    ResolutionRequest req = new ResolutionRequest(baseHandle,null,null,null);
            req.setSupportedProtocolVersion(thisSite);
			AbstractResponse response = resolver.sendRequestToServer(req, thisSite, thisSite.determineServer(baseHandle));

			if(response.responseCode==AbstractMessage.RC_HANDLE_NOT_FOUND) {
				return null;
			} else  if (response instanceof ErrorResponse){
				String msg = Util.decodeString( ((ErrorResponse)response).message );

				throw new HandleException(HandleException.INTERNAL_ERROR,
						AbstractMessage.getResponseCodeMessage(response.responseCode)+": "+msg);
			}

			values = ((ResolutionResponse)response).values;
		}
		else {
			values = storage.getRawHandleValues(caseSensitive ? baseHandle : Util.upperCase(baseHandle), null, null);
		}

		// allowing template handles to be constructed even without a base...
		if(values==null) {
			if (ns==null) return null;
			values = new byte[0][];
		}

		HandleValue[] handleValues = new HandleValue[values.length];
		for (int i = 0; i < values.length; i++) {
			handleValues[i] = new HandleValue();
			Encoder.decodeHandleValue(values[i],0,handleValues[i]);
		}
		NamespaceInfo thisNs = Util.getNamespaceFromValues(handleValues);
		if(thisNs!=null) {
			thisNs.setParentNamespace(ns);
			ns = thisNs;
		}
		
		if(ns==null || ns.getInheritedTag(NamespaceInfo.TEMPLATE_TAG)==null) {
			if(allValues) return values;
			ArrayList res = new ArrayList(values.length);
			for(int i = 0; i<values.length; i++) {
				byte[] clumpType = Encoder.getHandleValueType( values[i], 0 );
				int clumpIndex = Encoder.getHandleValueIndex( values[i], 0 );

				if( Util.isParentTypeInArray( typeList, clumpType) ||
						Util.isInArray( indexList, clumpIndex)) {
					res.add(values[i]);
				}
			}
			return (byte[][])res.toArray(new byte[res.size()][]);
		}

		HandleValue[] resVals = ns.templateConstruct(handleValues,inHandleStr,baseHandleStr,afterDelimStr,caseSensitive,resolver,recursionCount);
		if(resVals==null) return null;
		ArrayList res = new ArrayList(resVals.length);
		for(int i = 0; i < resVals.length; i++) {
			if( allValues || Util.isParentTypeInArray( typeList, resVals[i].getType()) ||
					Util.isInArray( indexList, resVals[i].getIndex())) {
				byte[] buf = new byte[Encoder.calcStorageSize(resVals[i])];
				Encoder.encodeHandleValue(buf,0,resVals[i]);
				res.add(buf);
			}
		}
		return (byte[][])res.toArray(new byte[res.size()][]);
	}




	/****************************************************************************
	 * Thread that purges old challenges after a specified period of time
	 * ////Currently hard-coded to 30 seconds - this should be configurable!!!!
	 ****************************************************************************/
	private class ChallengePurgeThread
	extends Thread
	{
		public void run() {
			while(keepRunning) {
				try {
					Thread.sleep(30000);
					long now = System.currentTimeMillis();
					for(IntTable.IntTableEnumerator enumeration = pendingAuthorizations.keys();
					enumeration.hasMoreElements(); ) {
						int key = enumeration.nextKey();
						ChallengeResponseInfo cri =
							(ChallengeResponseInfo)pendingAuthorizations.get(key);
						if(cri==null || cri.hasExpired()) {
							pendingAuthorizations.remove(key);
						}
					}

				} catch (Throwable e) {
					main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
							"Error purging pending authentications: "+e);
				}
			}
		}
	}

	//check whether the session id indicate a pending challenge
	private boolean pendingChallenge(int sessionId){
		if (sessionId < 0) return false;
		ChallengeResponseInfo crInfo =
			(ChallengeResponseInfo)pendingAuthorizations.get(sessionId);
		if (crInfo != null) return true;
		else return false;
	}

	private void sendResponse(ResponseMessageCallback callback, AbstractResponse response)
	throws HandleException
	{
	    if(thisSite.isPrimary) response.authoritative = true;
	    
		response.siteInfoSerial = thisSite.serialNumber;

		// If this is a certified request, sign the response.
		// We could be acting as a cache, so only sign the response
		//  if the cacheCertify flag is set, or if the response
		//  was handled locally (ie there is no other signature
		//  on it).
		if(response.certify && (response.cacheCertify || response.signature==null)){

			boolean signed = false;

			//session mannered response signing: use session key, return MAC code
			//with the message
			if (response.sessionId > 0) {
				// try to use session key to certify first ...  in case of
				// OC_SESSION_EXCHANGEKEY and OC_SESSION_SETUP, the client need to wait
				// for the response to confirm the session key is established.  so
				// don't sign with the session key if the request is
				// OC_SESSION_EXCHANGEKEY or OC_SESSION_SETUP.
				ServerSideSessionInfo sssinfo = getSession(response.sessionId);

				if (sssinfo != null &&
						response.opCode != AbstractMessage.OC_SESSION_SETUP &&
						response.opCode != AbstractMessage.OC_SESSION_EXCHANGEKEY) {
					try {
					    response.sessionCounter = sssinfo.getNextSessionCounter();
						response.signMessage(sssinfo.getSessionKey());
						signed = true;
					}
					catch (Exception e) {
						main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
								"Exception signing response: "+e);
						signed = false;
					}
				}
			}

			if (!signed) { //if session key signing fails
				// normal way of sending certified response through signature of server
				// private key
				try {
					// rotating through an array of signatures avoids a bottleneck
					// since we have to 'synchronize' on the signature while signing
					int sigIndex = currentSigIndex++;
					if(sigIndex >= serverSignatures.length)
						sigIndex = currentSigIndex = 0;

					Signature sig = serverSignatures[sigIndex];

					synchronized(sig) {
						response.signMessage(sig);
					}
				} catch (Exception e) {
					// If we get an error while signing the response, we return
					// the unsigned message anyway.  Maybe an error message would
					// be better?
					main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
							"Exception signing response: "+e);
				}
			}
		}

		callback.handleResponse(response);
	}

	public void disable() {
	    serverEnabled = false;
	}

	public void enable() {
	    serverEnabled = true;
	}



	/**********************************************************************
	 * Given a request object, handle the request and return a response
	 **********************************************************************/
	public void processRequest(AbstractRequest req, ResponseMessageCallback callback)
	throws HandleException
	{
		if(!serverEnabled) {
			sendResponse(callback, new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
					MSG_SERVER_TEMPORARILY_DISABLED));
			return;
		}

		// handle the request...
		processRequest(req, null, null, callback);
	}


	private void processRequest(AbstractRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq,
			ResponseMessageCallback callback)
	throws HandleException
	{
	        numRequests.incrementAndGet();

		// if the session id doesn't indicate a challenge entry or a valid session entry
		// then it is a invalid session
		if (req.sessionId > 0 && !pendingChallenge(req.sessionId) && !validSession(req)){
			sendResponse(callback,new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
					MSG_INVALID_SESSION_OR_TIMEOUT));
			return;
		}

		// please take note that all the req.authInfo is null in server side (this method)
		// even for administrative request.
		// this is because the decode message process can't re-produce the authInfo
		// on server side.
		// please see Encoder.java, all decodeMessage methods.

		// insert the request id into session information if the request has a valid session link
		validSession(req);

		// see if we need to be primary
		if(req.isAdminRequest && !isPrimary) {
			sendResponse(callback,
					new ErrorResponse(req, AbstractMessage.RC_SERVER_NOT_RESP,
							MSG_NOT_A_PRIMARY));
			return;
		}

		// now process the request
		// TODO: Properly increment request count of each type.

		switch(req.opCode) {
		case AbstractMessage.OC_GET_SITE_INFO:
			sendResponse(callback, new GetSiteInfoResponse(req, thisSite));
			break;
		case AbstractMessage.OC_RESOLUTION:
			// handle a resolution request-
		        numResolutionRequests.incrementAndGet();
			sendResponse(callback, doResolution((ResolutionRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_LIST_HANDLES:
			doListHandles(callback, (ListHandlesRequest)req, cRes, crReq);
			break;
		case AbstractMessage.OC_SESSION_SETUP:
			sendResponse(callback, doSessionSetup((SessionSetupRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_SESSION_EXCHANGEKEY:
			sendResponse(callback, doKeyExchange((SessionExchangeKeyRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_SESSION_TERMINATE:
			sendResponse(callback, doSessionTerminate((GenericRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_RESPONSE_TO_CHALLENGE:
			// look up the challenge that matches this response in the
			// pendingAuthorizations table
			ChallengeResponseInfo crInfo =
				(ChallengeResponseInfo)pendingAuthorizations.get(req.sessionId);
			if(crInfo==null || crInfo.hasExpired()) {
				sendResponse(callback,
						new ErrorResponse(req, AbstractMessage.RC_AUTHEN_TIMEOUT, null));
			} else {
				processRequest(crInfo.originalRequest, crInfo.challenge,
						(ChallengeAnswerRequest)req, callback);
			}
			break;
		case AbstractMessage.OC_VERIFY_CHALLENGE:
			sendResponse(callback, verifyChallenge((VerifyAuthRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_DUMP_HANDLES:
			sendResponse(callback, doDumpHandles((DumpHandlesRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_ADD_VALUE:
		        numAdminRequests.incrementAndGet();
			sendResponse(callback, doAddValue((AddValueRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_REMOVE_VALUE:
		        numAdminRequests.incrementAndGet();
			sendResponse(callback, doRemoveValue((RemoveValueRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_MODIFY_VALUE:
                        numAdminRequests.incrementAndGet();
			sendResponse(callback, doModifyValue((ModifyValueRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_CREATE_HANDLE:
                        numAdminRequests.incrementAndGet();
			sendResponse(callback, doCreateHandle((CreateHandleRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_DELETE_HANDLE:
                        numAdminRequests.incrementAndGet();
			sendResponse(callback, doDeleteHandle((DeleteHandleRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_GET_NEXT_TXN_ID:
			sendResponse(callback, getNextTxnId((GenericRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_RETRIEVE_TXN_LOG:
			sendResponse(callback,
					doRetrieveTxnLog((RetrieveTxnRequest)req, cRes, crReq));
			break;
		case AbstractMessage.OC_HOME_NA:
			sendResponse(callback, doHomeNA(req, cRes, crReq));
			break;
		case AbstractMessage.OC_UNHOME_NA:
			sendResponse(callback, doUnhomeNA(req, cRes, crReq));
			break;
		case AbstractMessage.OC_BACKUP_SERVER:
			sendResponse(callback, doBackup((GenericRequest)req, cRes, crReq));
			break;
		default:
			throw new HandleException(HandleException.INTERNAL_ERROR,
					"Unknown operation: "+req.opCode);
		}
	}


	class ChallengeResponseInfo
	{
		long timeStarted;
		int sessionId;
		ChallengeResponse challenge;
		AbstractRequest originalRequest;

		final boolean hasExpired() {
			return timeStarted < (System.currentTimeMillis()-maxAuthTime);
		}
	}


	/************************************************************************
	 * Return a message with the next consecutive transaction identifier.
	 ************************************************************************/
	private final AbstractResponse getNextTxnId(GenericRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		return new NextTxnIdResponse(req, getNextTxnId());
	}

	/** Get the next transaction ID  */
	private final long getNextTxnId()
	throws HandleException
	{
		long thisTxnId;
		synchronized(NEXT_TXN_ID_LOCK) {

			// if there is not yet any nextTxnId, get one from the authoritative server
			// in this service.  This is done because not all upgraded servers maintain
			// their own next-transaction-ID since IDs had to be unique across the entire
			// site and not just within
			// transaction IDs be unique across the site - not just within a single
			// server.
			if(!nextTxnIdInitialized && thisSite.servers.length>1 &&
					thisSite.determineServerNum(Common.SERVER_TXN_ID_HANDLE)!=thisServerNum) {
				// if the txn ID lives on another server and we don't yet have one
				long tmpID = retrieveNextTxnId();
				nextTxnId = Math.max(nextTxnId+1, tmpID);
			} else {
				nextTxnId++;
			}
			nextTxnIdInitialized = true;

			if(nextTxnId < 0) nextTxnId = 0;
			thisTxnId = nextTxnId;
			try {
			    new AtomicFile(txnIdFile).writeFully(Util.encodeString(String.valueOf(nextTxnId)));
			} catch (Exception e) {
				throw new HandleException(HandleException.INTERNAL_ERROR,
				"Unable to store new transaction ID");
			}
		}
		return thisTxnId;
	}


	/************************************************************************
	 * Tells the server that it will now be responsible for handles under
	 * the specified naming authority.
	 ************************************************************************/
	private final AbstractResponse doHomeNA(AbstractRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		//if(!Util.startsWithCI(req.handle, Common.NA_HANDLE_PREFIX)) {
		//      main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
		//                    "Was asked to home non-naming authority handle: '"+
		//                    Util.decodeString(req.handle)+"' ");
		//  return new ErrorResponse(req, AbstractRequest.RC_ERROR,
		//                           MSG_NOT_A_NA_HANDLE);
		//}

		//create challenge when there is no challenge answer info, or it is not a valid session
		if( (cRes==null || crReq==null) && !authenticatedSession(req))
			return createChallenge(req);

		boolean hasPermission = false;
		Vector valuesToTraverse = new Vector();
		ValueReference thisAdmin = null;

		if (crReq != null) {
			thisAdmin = new ValueReference(crReq.userIdHandle,
					crReq.userIdIndex);
		} else {
			//a valid session is assumed
			ServerSideSessionInfo sssinfo = getSession(req.sessionId);
			if (sssinfo != null) {
				thisAdmin = new ValueReference(sssinfo.identityKeyHandle,
						sssinfo.identityKeyIndex);
			} else
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
						MSG_INVALID_SESSION_OR_TIMEOUT);
		}

		for(int i=0; i<serverAdmins.length; i++) {
			ValueReference admin = serverAdmins[i];
			if(admin.equals(thisAdmin)) {
				hasPermission = true;
				break;
			} else {
				valuesToTraverse.addElement(admin);
			}
		}

		if(!hasPermission) {
		    for(int i=0; i<homeOnlyAdmins.length; i++) {
		        ValueReference admin = homeOnlyAdmins[i];
		        if(admin.equals(thisAdmin)) {
		            hasPermission = true;
		            break;
		        } else {
		            valuesToTraverse.addElement(admin);
		        }
		    }
		}

		if(!hasPermission) {
			if(!isAdminInGroup(thisAdmin, valuesToTraverse, new Vector())) {
				return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
			}
		}

		// If we got to here, then they are in the admin list for this server
		// Now we just need to verify that they are who they claim to be.
		AbstractResponse verifyResp = verifyIdentity(cRes, crReq, req);
		if(verifyResp!=null) {
			return verifyResp;
		}
		// if this NA handle hashes to this server, make a transaction for this
		// operation
		if(thisSite.determineServerNum(req.handle)==thisServerNum) {
			try {
				if(!insertTransaction(req.handle, Transaction.ACTION_HOME_NA)) {
					main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
					"Unable to save HOME-NA transaction.");
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
						"Error committing transaction: "+e);
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
			}
		}

		try {
			storage.setHaveNA(req.handle, true);
		} catch (HandleException e) {
			main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
					"Unable to \"home\" naming authority \""+
					Util.decodeString(req.handle)+
					"\" after transaction was logged! - "+e);
			e.printStackTrace(System.err);
			if(e.getCode() == HandleException.STORAGE_RDONLY)
				return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
						MSG_SERVER_BACKUP);
			else
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
		}
		// add to resolver's local sites
		if(enableHomedPrefixNaLookupOptimization) {
		    SiteInfo ss[] = { thisSite };
		    resolver.getConfiguration().setLocalSites(new String(req.handle), ss);
		}
		// return success
		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}


	/************************************************************************
	 * Tells the server that it will *not* be responsible for handles under
	 * the specified naming authority any longer.
	 ************************************************************************/
	private final AbstractResponse doUnhomeNA(AbstractRequest req,
                                            ChallengeResponse cRes,
                                            ChallengeAnswerRequest crReq)
	throws HandleException
	{
		//if there are no challenge or challenge answer and there are no authenticated session,
		//create challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req) )
			return createChallenge(req);

		boolean hasPermission = false;
		Vector valuesToTraverse = new Vector();

		ValueReference thisAdmin = null;
		if (crReq != null && cRes != null) {
			thisAdmin = new ValueReference(crReq.userIdHandle,
					crReq.userIdIndex);
		}
		else {
			//an authenticated session is assumed. if not, time out or no session info available
			//use identityHandle, and index in session info
			ServerSideSessionInfo sssinfo = getSession(req.sessionId);
			if (sssinfo != null) {
				thisAdmin = new ValueReference(sssinfo.identityKeyHandle,
						sssinfo.identityKeyIndex);
			} else
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
						MSG_INVALID_SESSION_OR_TIMEOUT);
		}

		for(int i=0; i<serverAdmins.length; i++) {
			ValueReference admin = serverAdmins[i];
			if(admin.equals(thisAdmin)) {
				hasPermission = true;
				break;
			} else {
				valuesToTraverse.addElement(admin);
			}
		}

		if(!hasPermission) {
			if(!isAdminInGroup(thisAdmin, valuesToTraverse, new Vector())) {
				return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
			}
		}

		// If we got to here, then they are in the admin list for this server
		// Now we just need to verify that they are who they claim to be.
		AbstractResponse verifyResp = verifyIdentity(cRes, crReq, req);
		if(verifyResp!=null) {
			return verifyResp;
		}

		if(thisSite.determineServerNum(req.handle)==thisServerNum) {
			try {
				if(!insertTransaction(req.handle, Transaction.ACTION_UNHOME_NA)) {
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
						"Error committing transaction: "+e);
			}
		}

		try{
			storage.setHaveNA(req.handle, false);
		} catch (HandleException e) {
			main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
					"Unable to \"home\" naming authority \""+
					Util.decodeString(req.handle)+
					"\" after transaction was logged! - "+e);
            e.printStackTrace(System.err);
			if(e.getCode() == HandleException.STORAGE_RDONLY)
				return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
						MSG_SERVER_BACKUP);
			else
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
		}
		// remove from resolver's local sites
		if(enableHomedPrefixNaLookupOptimization) {
		    resolver.getConfiguration().setLocalSites(new String(req.handle), null);
		}
		// return success
		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}




	/************************************************************************
	 * retrieve the transaction logs starting from a certain date, for all
	 * of the handles that hash according to the hash specification in
	 * the RetrieveTxnRequest object.  This first authenticates the client
	 * based on the replication administrators listed in the configuration
	 * file.
	 ************************************************************************/
	private final AbstractResponse doRetrieveTxnLog(RetrieveTxnRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// XXX: should make sure there are transactions to replicate before
		//      checking admin?  or does that leak information?
		//
		// XXX: could probably refactor this authentication code to be shared
		//      with other methods.  would lessen the exposure area for security
		//      bugs.
		//
		// if there are no challenge or challenge answer, and there is no
		// authenticated session.
		//
		//create challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req)) {
			return createChallenge(req);
		}

		try {
			boolean hasPermission = false;
			Vector valuesToTraverse = new Vector();

			ValueReference thisAdmin = null;
			if (crReq != null && cRes != null) {
				thisAdmin = new ValueReference(crReq.userIdHandle,
						crReq.userIdIndex);
			}
			else {
				//an authenticated session is assumed
				ServerSideSessionInfo sssinfo = getSession(req.sessionId);
				if (sssinfo != null) {
					thisAdmin = new ValueReference(sssinfo.identityKeyHandle,
							sssinfo.identityKeyIndex);
				} else
					return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
							MSG_INVALID_SESSION_OR_TIMEOUT);
			}

			for(int i=0; i<replicationAdmins.length; i++) {
				ValueReference admin = replicationAdmins[i];
				if(admin.equals(thisAdmin)) {
					hasPermission = true;
					break;
				} else {
					valuesToTraverse.addElement(admin);
				}
			}

			if(!hasPermission) {
				// the requestor's user ID is not referred to in the
				// replication admins list, so check server admins
				for(int i=0; i<serverAdmins.length; i++) {
					ValueReference admin = serverAdmins[i];
					if(admin.equals(thisAdmin)) {
						hasPermission = true;
						break;
					} else {
						valuesToTraverse.addElement(admin);
					}
				}

				if(!isAdminInGroup(thisAdmin, valuesToTraverse, new Vector())) {
				    return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
				}
			}

			// If we got to here, then they are in the admin list for this NA
			// Now we just need to verify that they are who they claim to be.
			AbstractResponse verifyResp = verifyIdentity(cRes, crReq, req);
			if(verifyResp!=null) {
				return verifyResp;
			}
		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL, "Unable to authenticate retrieve txn req: "+e);
			return new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null);
		} finally {

			// after doing something big like retrieving transaction logs, we
			// should garbage collect to free up memory and close open files..
			try {
				System.gc();
				System.runFinalization();
			} catch (Throwable t) {}
		}

		// at this point, the replication admin has been authenticated,
		// so we can begin streaming the transactions.  If the requestor
		// is hopelessly out of date, they will be notified in the
		// streaming part of the message.
		String start = (new Date(req.lastQueryDate)).toString();
		String end = (new Date(System.currentTimeMillis())).toString();
		long count = nextTxnId - req.lastTxnId;
		if (count > 0 && nextTxnId > 0){
			String msg = "";
			if (count == 1){
				msg="Replicating 1 transaction from ["+start+"] to ["+end+"]";
			}
			else if (count > 1){
				msg="Replicating "+count+" transactions from ["+start+"] to ["+end+"]";
			}
			else {
				msg = "Replicating all transactions";
			}
			main.logError(ServerLog.ERRLOG_LEVEL_INFO, msg);
		}


		return new RetrieveTxnResponse(txnQueue, nextTxnId, req, false, storage, privateKey);
	}



	/************************************************************************
	 * retrieve the entire contents of the handle database that hash
	 * according to the hash specification in the DumpHandlesRequest.
	 * This first authenticates the client based on the replication
	 * administrators listed in the configuration file.
	 ************************************************************************/
	private final AbstractResponse doDumpHandles(DumpHandlesRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		//if there are no challenge or challenge answer, and there is no authenticated session,
		//create the challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req)) {
			return createChallenge(req);
		}

		try {
			boolean hasPermission = false;
			Vector valuesToTraverse = new Vector();

			ValueReference thisAdmin = null;

			if (crReq != null && cRes != null) {
				thisAdmin = new ValueReference(crReq.userIdHandle,
						crReq.userIdIndex);
			}
			else {
				//an authenticated session is assumed
				ServerSideSessionInfo sssinfo = getSession(req.sessionId);
				if (sssinfo != null) {
					thisAdmin = new ValueReference(sssinfo.identityKeyHandle,
							sssinfo.identityKeyIndex);
				} else
					return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
							MSG_INVALID_SESSION_OR_TIMEOUT);
			}

			for(int i=0; i<replicationAdmins.length; i++) {
				ValueReference admin = replicationAdmins[i];
				if(admin.equals(thisAdmin)) {
					hasPermission = true;
					break;
				} else {
					valuesToTraverse.addElement(admin);
				}
			}

			if(!hasPermission) {
				// the requestor's user ID is not referred to in the
				// replication admins list, so check server admins
				for(int i=0; i<serverAdmins.length; i++) {
					ValueReference admin = serverAdmins[i];
					if(admin.equals(thisAdmin)) {
						hasPermission = true;
						break;
					} else {
						valuesToTraverse.addElement(admin);
					}
				}

				if(!isAdminInGroup(thisAdmin, valuesToTraverse, new Vector())) {
					return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
				}
			}

			// If we got to here, then they are in the admin list for this NA
			// Now we just need to verify that they are who they claim to be.
			AbstractResponse verifyResp = verifyIdentity(cRes, crReq, req);
			if(verifyResp!=null)
				return verifyResp;
		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,"Unable to authenticate dump handles request: "+e);
			return new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null);
		}

		// at this point, the replication admin has been authenticated,
		// and the requestor is reasonably up to date, so we can begin
		// streaming the handles.
		return new DumpHandlesResponse(req, storage, txnQueue, replicationDaemon, privateKey);
	}


	/************************************************************************
	 * verify whether or not the response to the challenge given in the
	 * request object was "signed" using the specified user's secret key.
	 * Note: currently the ChallengeResponse and ChallengeAnswerRequest
	 * objects are ignored here.  The user ID, challenge and response
	 * are all contained in the VerifyAuthRequest object.
	 ************************************************************************/
	private final AbstractResponse verifyChallenge(VerifyAuthRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		AbstractResponse res = errorIfNotHaveHandle(req);
		if (res!=null) return res;
		int handleIndex = req.handleIndex;
		byte clumps[][] = getRawHandleValuesWithTemplate(req.handle,
				handleIndex==0 ? null : new int[] { req.handleIndex },
						handleIndex==0 ? Common.MD5_SECRET_KEY_TYPES : null,req.recursionCount);
		if(clumps==null || clumps.length<=0) {
			return new VerifyAuthResponse(req, false);
		}

		// Get the encoding type (hash algorithm) of the signature
		// and use that to get the signature.
		byte authSignature[];
		byte digestAlg = req.signedResponse[0];

		boolean oldFormat = !req.hasEqualOrGreaterVersion(2,1); 

		if(oldFormat) { // the old format - just the md5 digest, no length or alg ID
			digestAlg = Common.HASH_CODE_MD5;
			authSignature = req.signedResponse;
		} else { // new format w/ one byte md5 or sha1 identifiers
			switch(digestAlg) {
			case Common.HASH_CODE_MD5:
				authSignature = new byte[Common.MD5_DIGEST_SIZE];
				System.arraycopy(req.signedResponse, 1, authSignature, 0, Common.MD5_DIGEST_SIZE);
				break;
			case Common.HASH_CODE_SHA1:
				authSignature = new byte[Common.SHA1_DIGEST_SIZE];
				System.arraycopy(req.signedResponse, 1, authSignature, 0, Common.SHA1_DIGEST_SIZE);
				break;
			default:
				// could be an invalid hash type, but for now we'll assume that it is
				// just the old format - no hash type, just 16 bytes of md5
				////// this should be changed to throw an exception once the old
				////// clients are phased out.
				authSignature = req.signedResponse;
			digestAlg = Common.HASH_CODE_MD5;
			/**
             throw new HandleException(HandleException.MESSAGE_FORMAT_ERROR,
             "Invalid hash type in secret key signature: "+((int)digestAlg));
			 */
			}
		}


		HandleValue secretKeyValue = new HandleValue();
		for(int valIdx=0; valIdx<clumps.length; valIdx++) {
			Encoder.decodeHandleValue(clumps[valIdx], 0, secretKeyValue);
			if(handleIndex!=0 && handleIndex!=secretKeyValue.getIndex())
				continue;
			if(!secretKeyValue.hasType(Common.STD_TYPE_HSSECKEY))
				continue;

			byte realSignature[] =
				Util.doDigest(digestAlg, secretKeyValue.getData(), req.nonce,
						req.origRequestDigest, secretKeyValue.getData());

			if(realSignature!=null && realSignature.length>0 && Util.equals(realSignature, authSignature)) {
				return new VerifyAuthResponse(req, true);
			}
		}

		return new VerifyAuthResponse(req, false);
	}


	/************************************************************************
	 * process a delete handle request, with the specified authentication
	 * info (if any).
	 ************************************************************************/
	private final AbstractResponse doDeleteHandle(DeleteHandleRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{

		// make sure that we are supposed to take this handle...
		if( thisSite.servers.length > 1 &&
				thisSite.determineServerNum(req.handle)!=thisServerNum ) {
			return new ErrorResponse(req, AbstractMessage.RC_SERVER_NOT_RESP,
					MSG_WRONG_SERVER_HASH);
		}

		// we might want to check to see if we manage this naming authority...
		// I left this out because we should be able to delete handles
		// that are in the database by mistake.

		// make sure the user has authenticated.  If not, send a challenge
		if( (cRes==null || crReq==null) && !authenticatedSession(req) ) {
			return createChallenge(req);
		}

		// check to see if the given user has permission to delete this
		try {
			// get the admin records from the handle, also check to make sure
			// that the handle exists.
			byte clumps[][] = storage.getRawHandleValues((caseSensitive ?
					req.handle :
						Util.upperCase(req.handle)),
						null, Common.ADMIN_TYPES);
			if(clumps==null) {
				return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
			}

			AbstractResponse authResp = authenticateUser(req, cRes, crReq, DEL_HANDLE_PERM, clumps);
			if(authResp!=null) {
				return authResp;
			}

		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,"Unable to authenticate delete handle request: "+e);
			return new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null);
		}


		try {
			if(!insertTransaction(req.handle, Transaction.ACTION_DELETE_HANDLE)) {
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
			}

			storage.deleteHandle((caseSensitive ?
					req.handle :
						Util.upperCase(req.handle)));
		} catch(HandleException e) {
			main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
					"Error committing transaction: "+e);
			switch(e.getCode()) {
			case HandleException.HANDLE_DOES_NOT_EXIST:
				return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
			case HandleException.STORAGE_RDONLY:
				return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
						MSG_SERVER_BACKUP);
			default:
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
			}
		}

		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}


	/** returns the offset of the value with the given index, or
	 * -1 if there is no value with that index. */
	private static final int getHVByIndex(HandleValue values[], int index) {
		for(int i=0; i<values.length; i++) {
			if(values[i]!=null && values[i].getIndex()==index)
				return i;
		}
		return -1;
	}




	/************************************************************************
	 * process a remove handle-value request, with the specified
	 * authentication info (if any).
	 ************************************************************************/
	private final AbstractResponse doRemoveValue(RemoveValueRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// make sure that we are supposed to take this handle...
		AbstractResponse res = errorIfNotHaveHandle(req);
		if (res!=null) return res;

		// make sure the user has authenticated.  If not, send a challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req)) {
			return createChallenge(req);
		}

		byte handle[] = (caseSensitive ? req.handle : Util.upperCase(req.handle));

		synchronized(getWriteLock(req.handle)) {

			// get the current values...
			byte rawValues[][] =
				storage.getRawHandleValues(handle, null, null);
            if(rawValues==null) {
                return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
            }

			// decode the current values
			HandleValue values[] = new HandleValue[rawValues.length];
			for(int i=0; i<rawValues.length; i++) {
				values[i] = new HandleValue();
				Encoder.decodeHandleValue(rawValues[i], 0, values[i]);
			}

			// find the values to be removed, and the permissions needed to remove them
			// if the requestor supplied a non-existent index, then it will return an error.
			boolean needsRemAdminPerm = false;
			boolean needsRemValuePerm = false;
			int toBeRemovedIdxs[] = new int[req.indexes.length];
			for(int i=0; i<toBeRemovedIdxs.length; i++) {
				toBeRemovedIdxs[i] = getHVByIndex(values, req.indexes[i]);
				if(toBeRemovedIdxs[i]<0) {
					return new ErrorResponse(req, AbstractMessage.RC_VALUES_NOT_FOUND, null);
				}
				HandleValue val = values[toBeRemovedIdxs[i]];
				if(val.hasType(Common.ADMIN_TYPE))
					needsRemAdminPerm = true;
				else
					needsRemValuePerm = true;
			}

			int neededPerms[];
			if(needsRemValuePerm && needsRemAdminPerm)
				neededPerms = REM_ADM_AND_VAL_PERM;
			else if(needsRemAdminPerm)
				neededPerms = REM_ADM_PERM;
			else
				neededPerms = REM_VAL_PERM;

			// authenticate the user
			AbstractResponse authResp = authenticateUser(req, cRes, crReq, neededPerms, rawValues);
			if(authResp!=null) {
				return authResp;
			}

			// take the removed values out of the array
			int removed=0;
			for(int i=0; i<toBeRemovedIdxs.length; i++) {
				if(values[toBeRemovedIdxs[i]]!=null)
					removed++;
				values[toBeRemovedIdxs[i]] = null;
			}

			// copy the remaining values to a new array
			HandleValue newValues[] = new HandleValue[values.length-removed];
			int j=0;
			for(int i=0; i<values.length; i++) {
				if(values[i]!=null)
					newValues[j++] = values[i];
			}

			values = newValues;

			try {
				// log the transaction
				if(!insertTransaction(handle, Transaction.ACTION_UPDATE_HANDLE)) {
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}

				// save the updated values to the database
				storage.updateValue(handle, values);

			} catch(HandleException e) {
				main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
						"Error committing transaction: "+e);
				switch(e.getCode()) {
				case HandleException.HANDLE_DOES_NOT_EXIST:
					return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
				case HandleException.STORAGE_RDONLY:
					return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
							MSG_SERVER_BACKUP);
				default:
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}
			}
		}

		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}


	/** Combine the toAppend permissions with the current permissions.
	 * This assumes that the values of the permissions arrays are
	 * immutable. */
	private static final int[] combinePerms(int current[], int toAppend[]) {
		if(current==null) return toAppend;
		if(toAppend==null) return current;
		int newPerms[] = new int[current.length + toAppend.length];
		System.arraycopy(current, 0, newPerms, 0, current.length);
		System.arraycopy(toAppend, 0, newPerms, current.length, toAppend.length);
		return newPerms;
	}

	private final AbstractResponse doModifyValue(ModifyValueRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// make sure that we are supposed to take this handle...
		AbstractResponse res = errorIfNotHaveHandle(req);
		if (res!=null) return res;

		// make sure the user has authenticated.  If not, send a challenge
		// added checking up with session
		boolean isAnonymous = ((cRes==null || crReq==null)  && !authenticatedSession(req));
		boolean needsAuthentication = false;

		byte handle[] = (caseSensitive ? req.handle : Util.upperCase(req.handle));

		////////////////////////////////////////////////////////
		/////// begin synchronized if if-modified-since applies.
		////////////////////////////////////////////////////////

		synchronized(getWriteLock(req.handle)) {

			// get the current values...
			byte rawValues[][] =
				storage.getRawHandleValues(handle, null, null);
            if(rawValues==null) {
                return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
            }

			// decode all of the current values
			HandleValue values[] = new HandleValue[rawValues.length];
			for(int i=0; i<rawValues.length; i++) {
				values[i] = new HandleValue();
				Encoder.decodeHandleValue(rawValues[i],0,values[i]);
			}

			boolean needsModAdminPerm = false;
			boolean needsModValuePerm = false;
			boolean needsAdmToValPerm = false;
			boolean needsValToAdmPerm = false;

			boolean oldValIsAdmin;
			boolean newValIsAdmin;
			// find the indexes of the value to be modified.
			int toBeModifiedIdxs[] = new int[req.values.length];
			for(int i=0; i<toBeModifiedIdxs.length; i++) {
				toBeModifiedIdxs[i] = getHVByIndex(values, req.values[i].getIndex());
				if(toBeModifiedIdxs[i]<0) {
					return new ErrorResponse(req, AbstractMessage.RC_VALUES_NOT_FOUND, null);
				}
				HandleValue val = values[toBeModifiedIdxs[i]];
				if(!val.getAdminCanWrite()) {
					// if write access is not allowed by administrators, nobody can write!
					return new ErrorResponse(req, AbstractMessage.RC_INSUFFICIENT_PERMISSIONS,
							MSG_READ_ONLY_VALUE);
				}

				oldValIsAdmin = val.hasType(Common.ADMIN_TYPE);
				newValIsAdmin = req.values[i].hasType(Common.ADMIN_TYPE);

				if(isAnonymous) {
					if(!val.getAnyoneCanWrite()) {
						// if the requestor is anonymous but anonymous write access is not allowed
						return createChallenge(req);
					} else if(oldValIsAdmin || newValIsAdmin) {
						// no anonymous operations are allowed on admin records
						return createChallenge(req);
					}
				}

				if(oldValIsAdmin && newValIsAdmin)
					needsModAdminPerm = true;
				else if(oldValIsAdmin && !newValIsAdmin)
					needsAdmToValPerm = true;
				else if (!oldValIsAdmin && newValIsAdmin)
					needsValToAdmPerm = true;
				else if(!oldValIsAdmin && !newValIsAdmin && !val.getAnyoneCanWrite())
					needsModValuePerm = true;
			}

			int perms[] = null;
			if(needsModAdminPerm)
				perms = combinePerms(perms, MOD_ADM_PERM);
			if(needsModValuePerm)
				perms = combinePerms(perms, MOD_VAL_PERM);
			if(needsAdmToValPerm)
				perms = combinePerms(perms, ADM_TO_VAL_PERM);
			if(needsValToAdmPerm)
				perms = combinePerms(perms, VAL_TO_ADM_PERM);

			// authenticate the user...
			if(perms != null){
				AbstractResponse authResp = authenticateUser(req, cRes, crReq, perms, rawValues);
				if(authResp!=null) {
					return authResp;
				}
			}

			// replace the old values with the new ones, and update the timestamps
			int now = (int)(System.currentTimeMillis()/1000);
			for(int i=0; i<toBeModifiedIdxs.length; i++) {
				values[toBeModifiedIdxs[i]] = req.values[i];
				values[toBeModifiedIdxs[i]].setTimestamp(now);
			}

			// check for duplicate index values
			for(int i=0; i<values.length; i++) {
				for(int j=i+1; j<values.length; j++) {
					if(values[i].getIndex()==values[j].getIndex()) {
						return new ErrorResponse(req, AbstractMessage.RC_INVALID_VALUE,
								Util.encodeString("Index conflict for "+values[j].getIndex()));
					}
				}
			}

			try {
				// log the transaction
				if(!insertTransaction(handle, Transaction.ACTION_UPDATE_HANDLE)) {
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}

				// save the updated value to the database
				storage.updateValue(handle, values);
			} catch(HandleException e) {
				main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
						"Error committing transaction: "+e);
				switch(e.getCode()) {
				case HandleException.HANDLE_DOES_NOT_EXIST:
					return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
				case HandleException.STORAGE_RDONLY:
					return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
							MSG_SERVER_BACKUP);
				default:
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}
			}
		}
		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}


	/************************************************************************
	 * process an add handle-value request, with the specified authentication
	 * info (if any).
	 ************************************************************************/
	private final AbstractResponse doAddValue(AddValueRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// make sure that we are supposed to take this handle...
		AbstractResponse res = errorIfNotHaveHandle(req);
		if (res!=null) return res;

        for(int i=0; i<req.values.length; i++) {
            if(req.values[i].getIndex() < 0) {
                return new ErrorResponse(req, AbstractMessage.RC_INVALID_VALUE, MSG_INDEXES_MUST_BE_POSITIVE);
            }
            for(int j=i+1; j<req.values.length; j++) {
                if(req.values[i].getIndex()==req.values[j].getIndex()) {
                    return new ErrorResponse(req, AbstractMessage.RC_VALUE_ALREADY_EXISTS,
                            Util.encodeString("Index conflict for "+req.values[j].getIndex()));
                }
            }
        }

		// make sure the user has authenticated.  If not, send a challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req) ) {
			return createChallenge(req);
		}

		byte handle[] = (caseSensitive ? req.handle : Util.upperCase(req.handle));

		if(req.values==null || req.values.length<=0)
			return new ErrorResponse(req, AbstractMessage.RC_INVALID_VALUE,
					MSG_EMPTY_VALUE_LIST);

		boolean didOverwriteExisting = false;
		synchronized(getWriteLock(req.handle)) {
			// get the current values...
			byte rawValues[][] =
				storage.getRawHandleValues(handle, null, null);

			if(rawValues==null) {
                return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
			}

			boolean needsAddAdminPerm = false;
			boolean needsAddValuePerm = false;

			// see what permissions they need... either add-admin, add-value, or both
			for(int i=0; i<req.values.length; i++) {
				if(req.values[i].hasType(Common.ADMIN_TYPE))
					needsAddAdminPerm = true;
				else
					needsAddValuePerm = true;
			}

			int neededPerms[] = null;
			if(needsAddValuePerm && needsAddAdminPerm)
				neededPerms = ADD_ADM_AND_VAL_PERM;
			else if(needsAddAdminPerm)
				neededPerms = ADD_ADM_PERM;
			else
				neededPerms = ADD_VAL_PERM;

			// combine the to-be-added values with the current values
			HandleValue values[] = new HandleValue[rawValues.length + req.values.length];
			int i=0;
			for(i=0; i<rawValues.length; i++) {
				values[i] = new HandleValue();
				Encoder.decodeHandleValue(rawValues[i], 0, values[i]);
			}
			int now = (int)(System.currentTimeMillis()/1000);
			for(int j=0; j<req.values.length; j++) {
				req.values[j].setTimestamp(now);
				values[i+j] = req.values[j];
			}

			// check for duplicate index values
			int removed = 0;
			outerloop:
			for(i=0; i<rawValues.length; i++) {
				for(int j=rawValues.length; j<values.length; j++) {
					if(values[i].getIndex()==values[j].getIndex()) {
					    if(req.overwriteWhenExists) {
					        removed++;
					        didOverwriteExisting = true;
					        values[i] = null;
					        continue outerloop;
					    }
					    else {
					        return new ErrorResponse(req, AbstractMessage.RC_VALUE_ALREADY_EXISTS,
								Util.encodeString("Index conflict for "+values[j].getIndex()));
					    }
					}
				}
			}

			if(removed>0) {
			    HandleValue[] fixedValues = new HandleValue[values.length - removed];
			    int index = 0;
			    for(i = 0; i < values.length; i++) {
			        if(values[i]!=null) fixedValues[index++] = values[i];
			    }
			    values = fixedValues;
			    if(neededPerms==ADD_VAL_PERM) neededPerms = ADD_ADM_AND_VAL_PERM;
			}

            // authenticate the user...
            AbstractResponse authResp = authenticateUser(req, cRes, crReq, neededPerms, rawValues);
            if(authResp!=null) {
                return authResp;
            }

			try {
				// log the transaction
				if(!insertTransaction(handle, Transaction.ACTION_UPDATE_HANDLE)) {
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}

				// save the updated value to the database
				storage.updateValue(handle, values);

			} catch(HandleException e) {
				main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
						"Error committing transaction: "+e);
				switch(e.getCode()) {
				case HandleException.HANDLE_DOES_NOT_EXIST:
					return new ErrorResponse(req, AbstractMessage.RC_HANDLE_NOT_FOUND, null);
				case HandleException.STORAGE_RDONLY:
					return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
							MSG_SERVER_BACKUP);
				default:
					return new ErrorResponse(req, AbstractMessage.RC_ERROR,
							MSG_INTERNAL_ERROR);
				}
			}

		}
		AbstractResponse resp = new GenericResponse(req, AbstractMessage.RC_SUCCESS);
		resp.overwriteWhenExists = didOverwriteExisting;
		return resp;
	}

	/**
	 * Used by doCreateHandle and doListHandle to retrieve the NA admin values.
	 * If allowNAAdmins is set to false in config.dct, it will always return an
	 * empty array.
	 *
	 * @param na The NA to retrieve admin values from.
	 *
	 * @return null on error, otherwise NA admin values.
	 **/
	private final byte[][] getNAAdminValues(byte na[])
	throws HandleException
	{
		if (!allowNAAdmins) return new byte[0][];

		ResolutionRequest authRequest =
			new ResolutionRequest(na, Common.ADMIN_TYPES, null, null);

		// this must be certified - oh yes!
		authRequest.certify = true;

		// get the admin records from the naming authority handle
		AbstractResponse response = resolver.processRequest(authRequest);

		if(response.responseCode==AbstractMessage.RC_HANDLE_NOT_FOUND) {
            // If we got here, we think we're responsible for this NA;
            // so give it an extra chance to really exist
		    authRequest.authoritative = true;
		    response = resolver.processRequest(authRequest);
		}

		if(response.getClass() != ResolutionResponse.class) {
			return null;
		}

		return ((ResolutionResponse)response).values;
	}

	/**
	 * Returns null if authentication and authorization was successful and
	 * an ErrorResponse otherwise.
	 *
	 * FIXME: Always returns RC_INVALID_ADMIN on error, even if it should
	 *        return RC_INSUFFICIENT_PERMISSIONS.
	 **/
	private final AbstractResponse authenticateUser(AbstractRequest req,
			ChallengeResponse challenge,
			ChallengeAnswerRequest response,
			int operationIDs[],
			byte values[][])
	throws HandleException
	{
	    // set up the identity handle and index here for two cases:
	    // 1. for Challenge Answer Request (in response),
	    //    the identity is response.userIdHandle and response.userIdIndex
	    // 2. for session information
	    //    the identity is in the session info identityHandle, identityIndex

	    // if we are allowing wide-open access, shortcut the authorization process
	    if(wideOpenAccessAllowed) return null;

	    byte[] identityHandle = null;
	    int   identityIndex = -1;
	    ServerSideSessionInfo sessionInfo = getSession(req.sessionId);

	    if (challenge != null && response != null) {
	        identityHandle = response.userIdHandle;
	        identityIndex = response.userIdIndex;
	    } else {
	        // there was no challenge/response, so get the client ID from the
	        // session information, if any
	        if (sessionInfo == null) {
	            return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
	                    MSG_INVALID_SESSION_OR_TIMEOUT);
	        }
	        identityHandle = sessionInfo.identityKeyHandle;
	        identityIndex = sessionInfo.identityKeyIndex;
	    }

	    // if the server has been configured to require sessions in order to perform
	    // administration then reject any requests that are not contained in a session
	    // and are not attempting to set up a session
	    if(requireSessions && sessionInfo==null &&
	            req.opCode!=AbstractMessage.OC_SESSION_SETUP) {
	        System.err.println("rejecting non-session request: "+req+
	                "; session: "+getSession(req.sessionId));
	        return new ErrorResponse(req, AbstractMessage.RC_INVALID_CREDENTIAL,
	                MSG_SESSION_REQUIRED);
	    }

	    boolean hasAdminAccess = false;
	    Vector valuesToTraverse = new Vector();
	    Vector valuesTraversed = new Vector();

	    if(serverAdminFullAccess) {
	        // the current client has not already been found in the list of admins, so check
	        // the server admins if serverAdminFullAccess flag is turned on in config file
	        for (int i=0; serverAdmins!=null && i<serverAdmins.length; i++) {
	            try {
	                if (Util.equalsPrefixCI(identityHandle, serverAdmins[i].handle) &&
	                        identityIndex == serverAdmins[i].index){
	                    hasAdminAccess = true;
	                    break;
	                } else {
	                    valuesToTraverse.addElement(serverAdmins[i]);
	                }
	            } catch (Exception e){
	                main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
	                        "Error checking for server admin: "+e);
	            }
	        }

	        try {
	            if(hasAdminAccess) {
	                // this admin is directly referenced as an administrator (ie no group resolution needed)
	                return verifyIdentity(challenge, response, req);

	            } else if(isAdminInGroup(new ValueReference(identityHandle, identityIndex),
	                    valuesToTraverse, valuesTraversed)) {
	                // the administrator is indirectly referenced as part of a group (HS_VLIST)
	                return verifyIdentity(challenge, response, req);

	            }
	        } catch (Exception e) {
	            main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
	                    "Error authenticating: "+e);
	            return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
	        }
	    }

	    // look for HS_ADMIN records that give the user permission to perform
	    // the requested operation
	    try {
	        HandleValue tmpValue = new HandleValue();
	        AdminRecord admin = new AdminRecord();

	        if (!allowNAAdmins) values = new byte[0][];

	        // look for admin values with the proper permissions
	        if (values != null) {
	            for(int i=0; i<values.length; i++) {
	                try {
	                    Encoder.decodeHandleValue(values[i], 0, tmpValue);
	                    if(!tmpValue.hasType(Common.ADMIN_TYPE)) {
	                        continue;
	                    }

	                    // extract the admin record from the handle value
	                    Encoder.decodeAdminRecord(tmpValue.getData(), 0, admin);

	                } catch (Exception e) {
	                    main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
	                            "Error decoding possible admin value: "+e);
	                    continue;
	                }
	                boolean adminRecordIsRelevant = true;
	                for(int p=0; p<operationIDs.length; p++) {
	                    if(!admin.perms[operationIDs[p]]) { // doesn't have permission
	                        adminRecordIsRelevant = false;
	                        break;
	                    }
	                }
	                if(!adminRecordIsRelevant) {
	                    continue;
	                }

	                // if this admin record refers directly to the requestor's user ID,
	                // then they are allowed access (provided their authentication
	                // can be verified).
	                if(Util.equalsPrefixCI(admin.adminId, identityHandle) &&
	                        admin.adminIdIndex==identityIndex) {

	                    // the given administrator handle/index was specifically listed as an
	                    // administrator
	                    hasAdminAccess = true;
	                    break;

	                } else {
	                    // this admin record didn't directly refer to the current user, but
	                    // it could be a reference to an admin group.
	                    valuesToTraverse.addElement(new ValueReference(admin.adminId,admin.adminIdIndex));
	                }
	            }
	        }
	    } catch (Throwable e) {
	        main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
	                "Error authenticating: "+e);
	    }



	    try {
	        if(hasAdminAccess) {
	            // this admin is directly referenced as an administrator (ie no group resolution needed)
	            return verifyIdentity(challenge, response, req);

	        } else if(isAdminInGroup(new ValueReference(identityHandle, identityIndex),
	                valuesToTraverse, valuesTraversed)) {
	            // the administrator is indirectly referenced as part of a group (HS_VLIST)
	            return verifyIdentity(challenge, response, req);

	        } else {
	            // the client is not referenced as an administrator, either directly or indirectly
	            if(DEBUG_AUTHENTICATION) {
	                HandleValue[] handleValues = null;
	                if(values!=null) {
	                    handleValues = new HandleValue[values.length];
	                    for(int i = 0; i < values.length; i++) {
	                        handleValues[i] = new HandleValue();
	                        Encoder.decodeHandleValue(values[i], 0, handleValues[i]);
	                    }
	                }
	                System.err.println(new Date() + " authenticateUser INVALID_ADMIN: " + identityIndex + ":" + Util.decodeString(identityHandle) + " session:" + req.sessionId + " handle:" + Util.decodeString(req.handle) + " ops:" + java.util.Arrays.toString(operationIDs) + " values:" + java.util.Arrays.toString(handleValues));
	            }
	            return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);

	        }
	    } catch (Exception e) {
	        main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
	                "Error authenticating: "+e);
	        return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
	    }
	}



	/************************************************************************
	 * process a create handle request, with the specified authentication
	 * info (if any).
	 ************************************************************************/
	private final AbstractResponse doCreateHandle(CreateHandleRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// make sure that we are supposed to take this handle...
		AbstractResponse res = errorIfNotHaveHandle(req);
		if (res!=null) return res;

		// make sure the handle is valid utf8
		if(!Util.isValidString(req.handle, 0, req.handle.length)) {
			return new ErrorResponse(req, AbstractMessage.RC_INVALID_HANDLE,
					MSG_INVALID_ENCODING);
		}

		// make sure the user has authenticated.  If not, send a challenge
		// if an authenticated session is found, don't send a challenge

		if((cRes==null || crReq==null)  && !authenticatedSession(req)) {
			return createChallenge(req);
		}

		// check for duplicate handle index values
		for(int i=0; i<req.values.length; i++) {
			if(req.values[i].getIndex() < 0) {
				return new ErrorResponse(req, AbstractMessage.RC_INVALID_VALUE, MSG_INDEXES_MUST_BE_POSITIVE);
			}
			for(int j=i+1; j<req.values.length; j++) {
				if(req.values[i].getIndex()==req.values[j].getIndex()) {
					return new ErrorResponse(req, AbstractMessage.RC_VALUE_ALREADY_EXISTS,
							Util.encodeString("Index conflict for "+req.values[j].getIndex()));
				}
			}
		}


        // process the new handle values
        int now = (int)(System.currentTimeMillis()/1000);
        for(int i=0; req.values!=null && i<req.values.length; i++) {
            req.values[i].setTimestamp(now);
        }

        // if this is a sub-naming authority handle, we get the admin information
        // from the parent naming authority.  Otherwise, we get the admin info
        // from the naming authority handle.
        boolean isSubNAHandle = Util.isSubNAHandle(req.handle);
        byte[][] naVals;
		try {
			byte na[] = isSubNAHandle ? Util.getParentNAOfNAHandle(req.handle)
					: resolver.getNAHandle(req.handle);
			naVals = getNAAdminValues(na);
			if (naVals == null){
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
				"Unable to find admin group while creating handle");
				//ss: allow null admin-values so that server-admin can be used for authentication...
				//return new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null);
			}

		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,"Error while creating handle: "+e);
			e.printStackTrace();
			return new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null);
		}

		boolean didOverwriteExisting = false;
		synchronized(getWriteLock(req.handle)) {

		    byte[][] rawValues = storage.getRawHandleValues((caseSensitive ? req.handle : Util.upperCase(req.handle)), null, Common.ADMIN_TYPES);
            if(!req.overwriteWhenExists && rawValues!=null) {
                return new ErrorResponse(req, AbstractMessage.RC_HANDLE_ALREADY_EXISTS, null);
            }
            didOverwriteExisting = rawValues!=null;

            // authenticate the user
            boolean authorized = false;
            if(rawValues!=null) {
                AbstractResponse authResp = authenticateUser(req, cRes, crReq, ADD_ADM_PERM, rawValues);
                authorized = authResp==null;
            }
            if(!authorized && rawValues!=null) {
                AbstractResponse authResp = authenticateUser(req, cRes, crReq, DEL_HANDLE_PERM, rawValues);
                if(authResp!=null) {
                    return authResp;
                }
            }
            if(!authorized) {
                AbstractResponse authResp =
                    authenticateUser(req, cRes, crReq,
                            isSubNAHandle? ADD_SUB_NA_PERM : ADD_HANDLE_PERM,
                                    naVals);
                if(authResp!=null) return authResp;
            }

            // add the new handle to the database if it isn't already in there
		    try {

		        if(!insertTransaction(req.handle, rawValues==null ? Transaction.ACTION_CREATE_HANDLE : Transaction.ACTION_UPDATE_HANDLE)) {
		            return new ErrorResponse(req, AbstractMessage.RC_ERROR,
		                    MSG_INTERNAL_ERROR);
		        }

		        if(rawValues==null) {
		            storage.createHandle((caseSensitive ?
		                    req.handle :
		                        Util.upperCase(req.handle)),
		                        req.values);
		        }
		        else {
                    storage.updateValue((caseSensitive ?
                            req.handle :
                                Util.upperCase(req.handle)),
                                req.values);
		        }
		    } catch(HandleException e) {
		        main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
		                "Error committing transaction: "+e);
		        switch(e.getCode()) {
		        case HandleException.HANDLE_ALREADY_EXISTS:
		            return new ErrorResponse(req, AbstractMessage.RC_HANDLE_ALREADY_EXISTS, null);
		        case HandleException.STORAGE_RDONLY:
		            return new ErrorResponse(req, AbstractMessage.RC_SERVER_BACKUP,
		                    MSG_SERVER_BACKUP);
		        default:
		            return new ErrorResponse(req, AbstractMessage.RC_ERROR,
		                    MSG_INTERNAL_ERROR);
		        }
		    }
		}

		AbstractResponse resp = new GenericResponse(req, AbstractMessage.RC_SUCCESS);
		resp.overwriteWhenExists = didOverwriteExisting;
		return resp;
	}

	private AbstractResponse errorIfNotHaveHandle(AbstractRequest req) throws HandleException {
		if( thisSite.servers.length > 1 &&
				thisSite.determineServerNum(req.handle)!=thisServerNum) {
			return new ErrorResponse(req, AbstractMessage.RC_SERVER_NOT_RESP,
					MSG_WRONG_SERVER_HASH);
		}
		boolean haveHandle = false;
		try {
			// currently not using: actually perform a lookup to find the NA handle.
			// haveHandle = storage.haveNA(resolver.getNAHandle(req.handle));
			// instead look for syntactic prefixes:
			if(Util.hasSlash(req.handle)) {
				int slash = req.handle.length - 1;
				byte[] naHandle = null;
				while(slash>=0) {
					if(req.handle[slash]==(byte)'/') {
						naHandle = new byte[slash];
						System.arraycopy(req.handle,0,naHandle,0,slash);
						if(storageHaveNa(naHandle)) {
						    haveHandle = true;
						    break;
						}
					}
					slash--;
				}
			} else {
				int dot = req.handle.length - 1;
				while(dot>=0) {
					if(req.handle[dot]==(byte)'.') {
						byte[] naHandle = new byte[dot];
						System.arraycopy(req.handle,0,naHandle,0,dot);
						if(storageHaveNa(naHandle)) {
							haveHandle = true;
							break;
						}
					}
					dot--;
				}
			}
		}
		catch (HandleException e) {
			return new ErrorResponse(req, AbstractMessage.RC_ERROR,
					MSG_INTERNAL_ERROR);
		}
		if(!haveHandle) {
			return new ErrorResponse(req, AbstractMessage.RC_SERVER_NOT_RESP,
					MSG_NA_NOT_HOMED_HERE);
		}
		return null;
	}

	/************************************************************************
	 * process a resolution request, with the specified authentication
	 * info (if any).
	 ************************************************************************/
	private final AbstractResponse doResolution(ResolutionRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// this is a hook to enable returning computed values (server status, etc)
		// from requests for certain handles
		if(enableStatusHandle && Util.equalsCI(req.handle, SERVER_STATUS_HANDLE)) {
			// construct and return a set of handles values containing the status of
			// the server
			ArrayList vals = new ArrayList();

			StringBuffer status = new StringBuffer();
			Runtime jr = Runtime.getRuntime();
			status.append("freemem=").append(jr.freeMemory());
			status.append("; totalmem=").append(jr.totalMemory());
			status.append("; maxmem=").append(jr.maxMemory());
			status.append("; runtime=").append(System.currentTimeMillis()-startTime);
			status.append("; numreqs=").append(numRequests);
			vals.add(new HandleValue(1, SERVER_STATUS_HDL_TYPE,
					Util.encodeString(status.toString())));

			if(monitorDaemon!=null) {
			    vals.add(new HandleValue(2, SERVER_STATUS_HDL_TYPE,
			            Util.encodeString(monitorDaemon.getStatusString())));
			}
			    
			byte valBytes[][] = new byte[vals.size()][];
			for(int i=0; i<valBytes.length; i++) {
				HandleValue val = (HandleValue)vals.get(i);
				valBytes[i] = new byte[Encoder.calcStorageSize(val)];
				Encoder.encodeHandleValue(valBytes[i], 0, val);
			}


			return new ResolutionResponse(req, req.handle, valBytes);
		}


		byte clumps[][] = null;
		AbstractResponse res = errorIfNotHaveHandle(req);
		if(res==null){
			try {
				clumps = getRawHandleValuesWithTemplate(
						req.handle,
						req.requestedIndexes, req.requestedTypes, req.recursionCount);
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
						String.valueOf(getClass())+": error getting values: "+e);
				e.printStackTrace(System.err);
				return new ErrorResponse(req, AbstractMessage.RC_ERROR,
						MSG_INTERNAL_ERROR);
			}
			if(clumps==null){
				return new ErrorResponse(req,AbstractMessage.RC_HANDLE_NOT_FOUND, null);
			} else if(clumps.length==0) {
				return new ErrorResponse(req,AbstractMessage.RC_VALUES_NOT_FOUND, null);
			} else {
				return checkReadAccess(req, clumps, cRes, crReq);
			}
		} else {
			// we don't have the handle, need to resolve it
			if(allowRecursiveQueries && req.recursive) {
				// if we're not responsible for this handle and the recursive
				// flag is set, we should attempt to resolve the handle for the
				// user...
				req.recursionCount++;
				if(req.recursionCount > RECURSION_LIMIT)
					return new ErrorResponse(req, AbstractMessage.RC_RECURSION_COUNT_TOO_HIGH, null);

				req.recursive = false;
				req.clearBuffers();

				try {
                    return resolver.processRequest(req);
				}
				catch (Exception e) {
					return HandleException.toErrorResponse(req,e);
				}
			} else {
				// the request was non-recursive, so we should just return an
				// error instead of resolving the handle ourselves.
				return res;
			}
		}

	}


	/************************************************************************
	 * process a list-handles request, with the specified authentication
	 * info (if any).
	 ************************************************************************/
	private final void doListHandles(ResponseMessageCallback callback,
			ListHandlesRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		// make sure the server allows the list handles permission
		if(!allowListHdls)
			sendResponse(callback, new ErrorResponse(req, AbstractMessage.RC_OPERATION_NOT_SUPPORTED, MSG_NEED_LIST_HDLS_PERM));

		// make sure that we are supposed to take this handle...
		if(!storageHaveNa(req.handle)) {
			sendResponse(callback, new ErrorResponse(req, AbstractMessage.RC_SERVER_NOT_RESP,
					MSG_NA_NOT_HOMED_HERE));
			return;
		}

		// make sure the user has authenticated.  If not, send a challenge
		if((cRes==null || crReq==null)  && !authenticatedSession(req)) {
			sendResponse(callback, createChallenge(req));
			return;
		}

		try {
			// authenticate the user
			byte vals[][] = getNAAdminValues(req.handle);
			AbstractResponse authResp =
				authenticateUser(req, cRes, crReq, LIST_HDLS_PERM, vals);
			if(authResp!=null) {
				sendResponse(callback, authResp);
				return;
			}
		} catch (Exception e) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,"Auth error on list-handles request: "+e);
			sendResponse(callback, new ErrorResponse(req, AbstractMessage.RC_AUTHEN_ERROR, null));
			return;
		}

		// At this point, the requestor has authenticated as an admin
		// of the NA with permission to list handles.  We will now begin
		// sending back list handle responses...
		ListHandlesResponse response = new ListHandlesResponse(req, null);
		Enumeration listEnum = storage.getHandlesForNA(req.handle);
		byte handles[][] = new byte[LIST_HANDLES_PER_MSG][];
		int numHandles = 0;
		int numMessages = 0;
		boolean errorSendingResponse = false;
		boolean noHandles = true;
		while(!errorSendingResponse && listEnum.hasMoreElements()) {
			noHandles = false;
			handles[numHandles++] = (byte[])listEnum.nextElement();
			if(numHandles>=LIST_HANDLES_PER_MSG) {
				response.handles = handles;
				response.clearBuffers();
				response.continuous = listEnum.hasMoreElements();
				numHandles = 0;
				numMessages++;
				try {
					sendResponse(callback, response);
				} catch (Throwable t) {
					errorSendingResponse = true;
					System.err.println("Error sending response to list-handles request: "+t);
				}

				if(numMessages%100==0) {
				  System.gc();
				  Runtime rt = Runtime.getRuntime();
//				  System.err.println(" memory: free="+rt.freeMemory()+
//                             "; max="+rt.maxMemory()+
//				                     "; total="+rt.totalMemory());
				}
			}
		}

		if((noHandles || numHandles>0) && !errorSendingResponse) {
			byte tmpHandles[][] = new byte[numHandles][];
			System.arraycopy(handles, 0, tmpHandles, 0, numHandles);
			response.handles = tmpHandles;
			response.clearBuffers();
			response.continuous = false;
			numHandles = 0;
			sendResponse(callback, response);
		}
	}


	/***************************************************************************
	 * Check if the specified user has read access to the given handle values.
	 * If so, then return a resolution response.
	 * Otherwise, return a challenge to the user to authenticate.
	 ***************************************************************************/
	private final AbstractResponse checkReadAccess( ResolutionRequest req,
			byte clumps[][],
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		if(req.ignoreRestrictedValues) {
			// the request is only asking for the publicly readable values
			int numUnrestricted = 0;
			for(int i=0; i<clumps.length; i++) {
				if((Encoder.getHandleValuePermissions(clumps[i],0) &
						Encoder.PERM_PUBLIC_READ)!=0)
					numUnrestricted++;
			}

			if (numUnrestricted==0) return new ErrorResponse(req,AbstractMessage.RC_VALUES_NOT_FOUND, null);

			byte unrestrictedVals[][] = new byte[numUnrestricted][];
			numUnrestricted--;
			for(int i=clumps.length-1; i>=0; i--)
				if((Encoder.getHandleValuePermissions(clumps[i],0) &
						Encoder.PERM_PUBLIC_READ)!=0)
					unrestrictedVals[numUnrestricted--] = clumps[i];

			return new ResolutionResponse(req, req.handle, unrestrictedVals);
		}

		// if we got here, the user wants to be authenticated if necessary
		boolean needsauth = req.sessionId != 0;
		if(!needsauth) {
		    for(int i=0; i<clumps.length; i++) {
		        byte perms = Encoder.getHandleValuePermissions(clumps[i],0);
		        if((perms & Encoder.PERM_PUBLIC_READ)==0 ) {
		            if((perms & Encoder.PERM_ADMIN_READ)==0) {
		                // neither admin nor public read access is allowed...
		                return new ErrorResponse(req, AbstractMessage.RC_INSUFFICIENT_PERMISSIONS, null);
		            }
		            needsauth = true;
		            break;
		        }
		    }
		}

		if(!needsauth) {
			// no authentication needed...
			return new ResolutionResponse( req, req.handle, clumps);

		} else if((cRes==null || crReq==null) && !authenticatedSession(req)) {
			// needs authentication, none provided; send a challenge to authenticate
			return createChallenge(req);
		} else {
			// needs authentication, some provided
			byte adminClumps[][] = getRawHandleValuesWithTemplate(req.handle,
					Common.ADMIN_INDEXES,
					Common.ADMIN_TYPES,req.recursionCount);
			AbstractResponse authResp = authenticateUser(req, cRes, crReq,
					READ_VAL_PERM, adminClumps);
			if(authResp!=null)
				return authResp;
			else
				return new ResolutionResponse(req, req.handle, clumps);
		}
	}


	/*****************************************************************************
	 *
	 *  Get the next transaction ID.  If we are unable to get the next transaction ID
	 *  (for example, if the txn ID server is not responding) then we throw an exception.
	 *
	 */

	private final long retrieveNextTxnId()
	throws HandleException
	{
		// we need to get the transaction id from the transaction server.
		long nextTxnId = -1;
		synchronized (TRANSACTION_LOCK) {
			try {
				AbstractResponse resp = resolver.sendRequestToSite(nextTxnIdRequest, thisSite);
				if ((resp != null) && (resp.responseCode == AbstractMessage.RC_SUCCESS))
					nextTxnId = ((NextTxnIdResponse)resp).nextTxnId;
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_REALBAD, "Unable to acquire next transaction ID: " + e);
				if(e instanceof HandleException) throw (HandleException)e;
				throw new HandleException(HandleException.INTERNAL_ERROR,
						"Unable to acquire next transaction ID: "+e);
			}
		}

		if(nextTxnId<0) {
			main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
			"Unable to acquire next transaction ID");
			throw new HandleException(HandleException.INTERNAL_ERROR,
			"Unable to acquire next transaction ID");
		}
		return nextTxnId;
	}


	private int replicationPriority;

	/** Sets the replication priority for transactions on this server */
	public void setReplicationPriority(int i) {
	    replicationPriority = i;
	}

    private final boolean insertTransaction(byte handle[], byte action) {
        return insertTransaction(handle,action,0);
    }
	
    // if date is zero, we set it to current time, and also set the date in the replicationDb
    // if date is non-zero, this is a copied foreign transaction, and we assume replicationDb is already dealt with.
	public final boolean insertTransaction(byte handle[], byte action, long date) {
	    // first, we need to get the transaction id from the transaction server.
	    long nextTxnId;
	    synchronized (TRANSACTION_LOCK) {
	        if(date <= 0) {
	            date = System.currentTimeMillis();
	            if(replicationDaemon!=null) {
	                try {
	                    switch(action) {
	                    case Transaction.ACTION_CREATE_HANDLE:
	                    case Transaction.ACTION_UPDATE_HANDLE:
	                    case Transaction.ACTION_DELETE_HANDLE:
	                        date = replicationDaemon.adjustAndSetLastCreateOrDeleteDate(handle,date,replicationPriority);
	                        break;
	                    case Transaction.ACTION_HOME_NA:
	                    case Transaction.ACTION_UNHOME_NA:
	                        date = replicationDaemon.adjustAndSetLastHomeOrUnhomeDate(handle,date,replicationPriority);
	                        break;
	                    }
	                }
	                catch (Throwable e) {
	                    main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Unable to save transation in replication database");
	                    e.printStackTrace();
	                }
	            }
	        }
	        try {
	            nextTxnId = getNextTxnId();
	        } catch (Throwable e) {
	            main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
	                    "Unable to acquire next transaction ID: "+e);
	            e.printStackTrace();
	            return false;
	        }

	        if(nextTxnId<0) {
	            main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
	            "Unable to acquire next transaction ID");
	            return false;
	        }

	        try {
	            txnQueue.addTransaction(nextTxnId, handle, action, date);
	        } catch (Throwable e) {
	            main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,
	            "Unable to insert transaction into queue");
	            e.printStackTrace();
	            return false;
	        }
	    }
	    return true;
	}

	private final AbstractResponse createChallenge(AbstractRequest req)
	throws HandleException
	{
		ChallengeResponseInfo cri = new ChallengeResponseInfo();
		cri.timeStarted = System.currentTimeMillis();
		cri.challenge = new ChallengeResponse(req);
		cri.originalRequest = req;

		/*
    synchronized (pendingAuthorizations) {
      cri.sessionId = ++nextAuthId;
      cri.challenge.sessionId = cri.sessionId;
      pendingAuthorizations.put(cri.sessionId, cri);
    }
		 */
		synchronized (pendingAuthorizations) {

			//use session id if the req is in a valid session
			if (validSession(req))
				cri.sessionId = req.sessionId;
			else
				cri.sessionId = getNextSessionId();

			cri.challenge.sessionId = cri.sessionId;
			Object formerCri = pendingAuthorizations.put(cri.sessionId, cri);
			if(formerCri!=null) {
			    // already was a pending auth in this session
			    try {
			        main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,"Warning: clobbering pending authorization of " + ((ChallengeResponseInfo)formerCri).originalRequest);
			    } catch (Exception e) {}
			}
		}

		return cri.challenge;
	}

	private static synchronized int getNextSessionId()
	{
		return ++nextAuthId;
	}

	/**
	 * Determines if the administrator handle/index combination included in the
	 * admin parameter is listed in a group (aka HS_VLIST) that is included in
	 * the valuesToTraverse Vector (of ValueReference objects).  ValueReferences
	 * from the valuesToTraverse Vector are resolved and the contents of any HS_VLIST
	 * values are added to the valuesToTraverse vector if they aren't already in
	 * valuesTraversed.  If an exact match for the admin ValueReference is found
	 * then this returns true.  Otherwise it returns false.
	 */
	private final boolean isAdminInGroup(ValueReference admin, Vector valuesToTraverse,
			Vector valuesTraversed)
	{
	    while(valuesToTraverse.size() > 0) {
	        ValueReference val = (ValueReference)valuesToTraverse.elementAt(0);
	        valuesToTraverse.removeElementAt(0);

	        if(valuesTraversed.contains(val)) {
	            continue;
	        }

	        valuesTraversed.addElement(val);

	        ResolutionRequest req = new ResolutionRequest(val.handle, null,
	                new int[]{ val.index }, null);
	        HandleValue groupValue = null;
	        try {
	            req.certify = true;
	            AbstractResponse response;
	            // if the group handle lives on this server, find it locally
	            // instead of talking to ourself via the network.
	            if(errorIfNotHaveHandle(req)==null) {
	                response = doResolution(req, null, null);
	            } else {
	                response = resolver.processRequest(req);
	            }

	            if(response.responseCode==AbstractMessage.RC_SUCCESS &&
	                    response.opCode==AbstractMessage.OC_RESOLUTION) {
	                ResolutionResponse resResponse = (ResolutionResponse)response;
	                groupValue = new HandleValue();

	                for(int v=0; v<resResponse.values.length; v++) {
	                    Encoder.decodeHandleValue(resResponse.values[v], 0, groupValue);
	                    if(!groupValue.hasType(Common.STD_TYPE_HSVALLIST))
	                        continue;
	                    ValueReference valuesInGroup[] =
	                            Encoder.decodeValueReferenceList(groupValue.getData(), 0);
	                    for(int i=0; i<valuesInGroup.length; i++) {
	                        if(admin.equals(valuesInGroup[i])) {
	                            return true;
	                        } else if(!valuesToTraverse.contains(valuesInGroup[i]) &&
	                                !valuesTraversed.contains(valuesInGroup[i])) {
	                            valuesToTraverse.addElement(valuesInGroup[i]);
	                        }
	                    }
	                }
	            }
	        } catch (Throwable e) {
	            System.err.println("Error trying to resolve possible group: "+e);
	            e.printStackTrace(System.err);
	        }
	    }

	    if(DEBUG_AUTHENTICATION) System.err.println("" + new Date() + " isAdminInGroup fails: " + admin + " " + valuesTraversed);
	    return false;
	}

	/************************************************************************
	 * Verify that the person who signed the response actually is who they
	 * claim to be, and that the given response is a reply to the given
	 * challenge.  Returns null if the requestor's identity was verified.
	 * Otherwise it returns the error message that should be sent back to
	 * the requestor.

	 * Now added verify identity from session info
	 ************************************************************************/
	private final AbstractResponse verifyIdentity(ChallengeResponse cRes,
			ChallengeAnswerRequest crReq,
			AbstractRequest origReq)
	throws HandleException
	{
		if (cRes == null && crReq == null && authenticatedSession(origReq)) {
			// no challenge/response and there apparently was a session established
			ServerSideSessionInfo sssinfo = getSession(origReq.sessionId);
			if (sssinfo == null) {
				return new ErrorResponse(origReq, AbstractMessage.RC_SESSION_TIMEOUT,
						MSG_INVALID_SESSION_OR_TIMEOUT);
			}

			// the signature of the origReq shall be a HS_MAC types credential
			if (origReq.signature == null || origReq.signature.length <= 0) {
				//send a challenage response
				return new ErrorResponse(origReq,AbstractMessage.RC_INVALID_CREDENTIAL,
						Util.encodeString("Session request missing MAC code."));
			}

			byte[] sessionKey = sssinfo.getSessionKey();
			try {
				if (sessionKey!=null && origReq.verifyMessage(sessionKey)) {
		            try {
		                sssinfo.addSessionCounter(origReq.sessionCounter);
		            } catch(HandleException e) {
		                if(e.getCode()==HandleException.DUPLICATE_SESSION_COUNTER) {
		                    return new ErrorResponse(origReq,AbstractMessage.RC_SESSION_MESSAGE_REJECTED,Util.encodeString(e.getMessage()));
		                } else throw e;
		            }
				    return null;
				}
			} catch (Exception e) {
				e.printStackTrace();
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
						"Error verifying session key:" +e);
			}
			return new ErrorResponse(origReq, AbstractMessage.RC_INVALID_SESSION_KEY,
					Util.encodeString("Session authentication failed."));
		}

		if(crReq!=null && Util.equals(crReq.authType, Common.MD5_SECRET_KEY_TYPE)) {
			// there was a response to a challenge and it was based on a secret key

			// we verify secret keys by asking the responsible server if the
			// given response (md5 hash of the secret_key+nonce+request_digest)
			// is valid for the given challenge.

			VerifyAuthRequest vaReq =
				new VerifyAuthRequest(crReq.userIdHandle, cRes.nonce,
						cRes.requestDigest, cRes.rdHashType,
						crReq.signedResponse, crReq.userIdIndex, null);

			vaReq.certify = true;
			vaReq.setSupportedProtocolVersion(cRes);
			
			// if the callers identity handle lives on this server, find it locally
			// instead of talking to ourself via the network.
			AbstractResponse response;
			if(errorIfNotHaveHandle(vaReq)==null){
				response = verifyChallenge(vaReq, null, null);
			} else {
				response = resolver.processRequest(vaReq);
			}
			
			if(response instanceof VerifyAuthResponse) {
				if(((VerifyAuthResponse)response).isValid){

					//if this is a valid session,
					//set the client authenticated flag to true.
					//for the first time request, MAC wasn't checked, so check here.
					if (validSession(origReq)) {

						setSessionAuthenticated(origReq, crReq, true);

						boolean MACpass = true;
						ServerSideSessionInfo ssinfo = getSession(origReq.sessionId);
						if (ssinfo != null && ssinfo.getSessionKey() != null) {
							try {
								MACpass = origReq.verifyMessage(ssinfo.getSessionKey());

							}
							catch (Exception e) {
								System.err.println("Error verifying the original request MAC code:" + e);
								MACpass = false;
							}
						}
						if (MACpass) {
						    try {
						        ssinfo.addSessionCounter(origReq.sessionCounter);
						    } catch(HandleException e) {
						        if(e.getCode()==HandleException.DUPLICATE_SESSION_COUNTER) {
						            return new ErrorResponse(origReq,AbstractMessage.RC_SESSION_MESSAGE_REJECTED,Util.encodeString(e.getMessage()));
						        } else throw e;
						    }
					        return null;
						}
						else return new ErrorResponse(origReq, AbstractMessage.RC_INVALID_SESSION_KEY,
								Util.encodeString("The session key authentication failed."));
					}

					return null;
                }
            }
			return new ErrorResponse(origReq, AbstractMessage.RC_AUTHENTICATION_FAILED, null);

		} else if(crReq!=null &&
				(Util.equals(crReq.authType, Common.STD_TYPE_HSDSAPUBKEY) ||
						Util.equals(crReq.authType, Common.PUBLIC_KEY_TYPE))) {
			// there was a response to a challenge and it was based on a public key

			// verify that the challenge response was signed by the private key
			// associated with the administrators public key.

			// first retrieve the public key (checking server signatures, of course)
			ResolutionRequest req = new ResolutionRequest(crReq.userIdHandle, null,
					new int[] {crReq.userIdIndex},
					null);
			req.certify = true;
			AbstractResponse response = null;
			// if the callers identity handle lives on this server, find it locally
			// instead of talking to ourself via the network.
			if(errorIfNotHaveHandle(req)==null) {
			    response = doResolution(req, null, null);
			} else {
			        response = resolver.processRequest(req);
			}

			if(response.getClass()==ResolutionResponse.class) {
				ResolutionResponse rresponse = (ResolutionResponse)response;
				HandleValue values[] = rresponse.getHandleValues();
				if(values==null || values.length < 1) {
	                if(DEBUG_AUTHENTICATION) main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"No values kills auth " + crReq.userIdIndex + ":" + Util.decodeString(crReq.userIdHandle) + ": " + response);
					return new ErrorResponse(origReq, AbstractMessage.RC_AUTHENTICATION_FAILED, null);
				}
				try {
					// get the algorithm used to sign
					int algNameLen = Encoder.readInt(crReq.signedResponse,0);

					int offset = 0;
					byte hashAlgId[] = Encoder.readByteArray(crReq.signedResponse, offset);
					offset += Encoder.INT_SIZE + hashAlgId.length;

					// get the actual bytes of the signature
					byte sigBytes[] = Encoder.readByteArray(crReq.signedResponse, offset);
					offset += Encoder.INT_SIZE + sigBytes.length;

					// decode the public key
					PublicKey pubKey = Util.getPublicKeyFromBytes(values[0].getData(), 0);

					boolean verified = false;

					// get a signature object based on the public key
					Signature sig =
						Signature.getInstance(Util.getSigIdFromHashAlgId(hashAlgId,
								pubKey.getAlgorithm()));
					sig.initVerify(pubKey);

					// verify the signature
					sig.update(cRes.nonce);
					sig.update(cRes.requestDigest);

					if(sig.verify(sigBytes)) {

						//if this is a valid session,
						//set the client authenticated flag to true.
						//for the first time request, MAC wasn't checked, so check here.
						if (validSession(origReq)) {

							setSessionAuthenticated(origReq, crReq, true);

							boolean MACpass = true;
							ServerSideSessionInfo ssinfo = getSession(origReq.sessionId);
							if (ssinfo != null && ssinfo.getSessionKey() != null) {
								try {
									MACpass = origReq.verifyMessage(ssinfo.getSessionKey());
								}
								catch (Exception e) {
									System.err.println("Error verifying the original request MAC code:" + e);
									MACpass = false;
								}
							}
							if (MACpass) {
					            try {
					                ssinfo.addSessionCounter(origReq.sessionCounter);
					            } catch(HandleException e) {
					                if(e.getCode()==HandleException.DUPLICATE_SESSION_COUNTER) {
					                    return new ErrorResponse(origReq,AbstractMessage.RC_SESSION_MESSAGE_REJECTED,Util.encodeString(e.getMessage()));
					                } else throw e;
					            }
					            return null;
							}
							else return new ErrorResponse(origReq, AbstractMessage.RC_INVALID_SESSION_KEY,
									Util.encodeString("The session key authentication failed."));
						}

						return null;
					} else {
					    if(DEBUG_AUTHENTICATION) main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Verify failure kills auth " + crReq.userIdIndex + ":" + Util.decodeString(crReq.userIdHandle));
						return new ErrorResponse(origReq, AbstractMessage.RC_AUTHENTICATION_FAILED, null);
					}
				} catch (Exception e) {
				    if(DEBUG_AUTHENTICATION) {
				        main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Exception kills auth " + crReq.userIdIndex + ":" + Util.decodeString(crReq.userIdHandle) +":");
				        e.printStackTrace();
				    }
					return new ErrorResponse(origReq, AbstractMessage.RC_AUTHENTICATION_FAILED, null);
				}
			} else {
			    if(DEBUG_AUTHENTICATION) main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Bad response kills auth " + crReq.userIdIndex + ":" + Util.decodeString(crReq.userIdHandle) + ": " + response);
				return new ErrorResponse(origReq, AbstractMessage.RC_AUTHENTICATION_FAILED, null);
			}
		} else {
			return new ErrorResponse(origReq, AbstractMessage.RC_INVALID_CREDENTIAL, null);
		}
	}


	/************************************************************************
	 * Backup the server's database.
	 * This first authenticates the client based on the server
	 * administrators listed in the configuration file.
	 ************************************************************************/
	private final AbstractResponse doBackup(GenericRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		//if the client is not authenticated, send a challenge
		if((cRes==null || crReq==null) && !authenticatedSession(req)) {
			return createChallenge(req);
		}

		boolean hasPermission = false;
		Vector valuesToTraverse = new Vector();
		ValueReference thisAdmin = null;

		if (crReq != null) {
			thisAdmin = new ValueReference(crReq.userIdHandle,
					crReq.userIdIndex);
		} else {
			//an authenticated session is preassumed
			ServerSideSessionInfo ssinfo = getSession(req.sessionId);
			if(ssinfo != null) {
				thisAdmin = new ValueReference(ssinfo.identityKeyHandle,
						ssinfo.identityKeyIndex);
			} else {
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
						MSG_INVALID_SESSION_OR_TIMEOUT);
			}
		}

		for(int i=0; i<backupAdmins.length; i++) {
			ValueReference admin = backupAdmins[i];
			if(admin.equals(thisAdmin)) {
				hasPermission = true;
				break;
			} else {
				valuesToTraverse.addElement(admin);
			}
		}

		if(!hasPermission) {
			// the requestor's user ID is not referred to in the
			// backup admins list, so check server admins
			for(int i=0; i<serverAdmins.length; i++) {
				ValueReference admin = serverAdmins[i];
				if(admin.equals(thisAdmin)) {
					hasPermission = true;
					break;
				} else {
					valuesToTraverse.addElement(admin);
				}
			}

			if(!isAdminInGroup(thisAdmin, valuesToTraverse, new Vector())) {
				return new ErrorResponse(req, AbstractMessage.RC_INVALID_ADMIN, null);
			}
		}

		// If we got to here, then they are in the admin list for this server
		// Now we just need to verify that they are who they claim to be.
		AbstractResponse verifyResp = verifyIdentity(cRes, crReq, req);
		if(verifyResp!=null) {
			return verifyResp;
		}

		try {
			storage.checkpointDatabase();
		} catch(Exception e){
			main.logError(ServerLog.ERRLOG_LEVEL_FATAL,
					"Error backup server: "+e);
			return new ErrorResponse(req, AbstractMessage.RC_ERROR,
					MSG_INTERNAL_ERROR);
		}

		return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
	}

	private final AbstractResponse doSessionTerminate(GenericRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException {

		//check client authentication later! use session info, for session is still available right now..
		ServerSideSessionInfo ssinfo = getSession(req.sessionId);
		if (ssinfo != null){

			boolean authenticated = false;
			try {
				authenticated = req.verifyMessage(ssinfo.getSessionKey());
			} catch (Exception e) {
				System.err.println(e.getMessage());
				authenticated = false;
			}

			if (!authenticated) {
				return new ErrorResponse(req, AbstractMessage.RC_INVALID_SESSION_KEY,
						Util.encodeString("Invalid session key."));
			}
			
			try {
			    ssinfo.addSessionCounter(req.sessionCounter);
			} catch(HandleException e) {
			    if(e.getCode()==HandleException.DUPLICATE_SESSION_COUNTER) {
			        return new ErrorResponse(req,AbstractMessage.RC_SESSION_MESSAGE_REJECTED,Util.encodeString(e.getMessage()));
			    } else throw e;
			}
			
			sessions.removeSession(req.sessionId);
			return new GenericResponse(req, AbstractMessage.RC_SUCCESS);
		}
		else {
			return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
					Util.encodeString("Can not get session info."));
		}
	}


	private final AbstractResponse doSessionSetup(SessionSetupRequest req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{
		SessionSetupResponse rsp = new SessionSetupResponse(req, null);

		PublicKey pubKey = null; // the exchange key for SessionInfo
		byte sessionKey[] = null; // the exchange key for SessionInfo
		boolean oldClient = !req.hasEqualOrGreaterVersion((byte)2, (byte)2);
		int sessionKeyAlg = oldClient ? HdlSecurityProvider.ENCRYPT_ALG_DES : encryptionAlgorithm;

		if (req.keyExchangeMode == Common.KEY_EXCHANGE_CIPHER_HDL) {
			try {
				HandleValue vals[] = resolver.resolveHandle(
						Util.decodeString(req.exchangeKeyHandle),
						null, new int[]{req.exchangeKeyIndex});
				HandleValue pubkeyval = null;
				for(int i=0; vals!=null && i<vals.length; i++) {
					if(vals[i].hasType(Common.PUBLIC_KEY_TYPE)) {
						pubkeyval = vals[i];
						break;
					}
				}
				if (pubkeyval==null) {
					main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
					"Error initializing session with hdl cipher: no key found.");
					return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
							Util.encodeString("No key found in key exchange handle."));
				}
				pubKey = Util.getPublicKeyFromBytes(vals[0].getData(), 0);
				sessionKey = HdlSecurityProvider.getInstance().generateSecretKey(sessionKeyAlg);
				byte encryptKey[] = Util.substring(sessionKey, Encoder.INT_SIZE);
				rsp.data = Util.encrypt(pubKey, oldClient ? encryptKey : sessionKey, rsp.majorProtocolVersion, rsp.minorProtocolVersion);
				sessionKey = encryptKey;
			} catch (Exception e){
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
						"Error initializing session with hdl cipher: "+e);
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
						Util.encodeString("Error performing hdl cipher key exchange."));
			}
		} else if (req.keyExchangeMode == Common.KEY_EXCHANGE_CIPHER_CLIENT){
			try {
				pubKey = Util.getPublicKeyFromBytes(req.publicKey, 0);
				sessionKey = HdlSecurityProvider.getInstance().generateSecretKey(sessionKeyAlg);
				byte encryptKey[] = Util.substring(sessionKey, Encoder.INT_SIZE);
				rsp.data = Util.encrypt(pubKey, oldClient ? encryptKey : sessionKey, rsp.majorProtocolVersion, rsp.minorProtocolVersion);
				sessionKey = encryptKey;
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
						"Error initializing client cipher session: "+e);
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
						Util.encodeString("Error performing client cipher key exchange."));

			}
		} else if (req.keyExchangeMode == Common.KEY_EXCHANGE_CIPHER_SERVER){
		    if(!(pubKey instanceof RSAPublicKey)) {
		        return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
		                Util.encodeString("KEY_EXCHANGE_CIPHER_SERVER not supported"));
		    }
		    try {
				rsp.data = Util.getBytesFromPublicKey(pubKey);
			} catch (Exception e) {
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
						"Error initializing server cipher session: "+e);
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
						Util.encodeString("Error initializing server cipher session: "+e));
			}
		} else if (req.keyExchangeMode == Common.KEY_EXCHANGE_DH) {
			try {
				pubKey = Util.getPublicKeyFromBytes(req.publicKey, 0);
				HdlSecurityProvider provider = HdlSecurityProvider.getInstance();
				DHPublicKey pub = (DHPublicKey)pubKey;
				DHParameterSpec dhSpec = pub.getParams();
				KeyPair kp = provider.generateDHKeyPair(dhSpec.getP(), dhSpec.getG());

				DHPrivateKey priv = (DHPrivateKey)kp.getPrivate();

				oldClient = !req.hasEqualOrGreaterVersion((byte)2, (byte)4);
				if(!oldClient) {
				    byte[] pubKeyBytes = Util.getBytesFromPublicKey(kp.getPublic());
				    rsp.data = new byte[Encoder.INT_SIZE + pubKeyBytes.length];
				    Encoder.writeInt(rsp.data,0,encryptionAlgorithm);
				    System.arraycopy(pubKeyBytes,0,rsp.data,Encoder.INT_SIZE,pubKeyBytes.length);
				    sessionKey = provider.getKeyFromDH(pub, priv, encryptionAlgorithm);
					sessionKey = Util.substring(sessionKey, Encoder.INT_SIZE);
					sessionKeyAlg = encryptionAlgorithm;
				} else {
	                rsp.data = Util.getBytesFromPublicKey(kp.getPublic());
					sessionKey = provider.getDESKeyFromDH(pub, priv);
					sessionKeyAlg = HdlSecurityProvider.ENCRYPT_ALG_DES;
				}
			} catch (Exception e){
				e.printStackTrace();
				main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
						"Error initializing DH session: "+e);
				return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
						Util.encodeString("Error encoding public session key"));
			}
		} else {
			return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
					Util.encodeString("Unrecognized key exchange mode"));
		}
		int sessionId = getNextSessionId();
		rsp.sessionId = sessionId;

		ServerSideSessionInfo sinfo =
			new ServerSideSessionInfo(sessionId, sessionKey, req.identityHandle,
					req.identityIndex, pubKey, req.keyExchangeMode, rsp.majorProtocolVersion, rsp.minorProtocolVersion);
		// set other attributes of the session
		sinfo.setEncryptionAlgorithmCode(sessionKeyAlg);
		sinfo.setTimeOut(req.timeout);
		sinfo.setEncryptedMesssageFlag(req.encryptAllSessionMsg);
		sinfo.setAuthenticateMessageFlag(req.authAllSessionMsg);
		sessions.addSession(sinfo);

		// store the request id of SESSION_SETUP_REQUEST into this modified session
		// info
		return rsp;
	}



	/**
	 * client has sent the session key to finish a KEY_EXCHANGE_CIPHER_SERVER.
	 **/
	private final AbstractResponse doKeyExchange(SessionExchangeKeyRequest  req,
			ChallengeResponse cRes,
			ChallengeAnswerRequest crReq)
	throws HandleException
	{

		if (!validSession(req)) {
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
			"Bad server-cipher key exchange request");
			return new ErrorResponse(req, AbstractMessage.RC_SESSION_TIMEOUT,
					MSG_INVALID_SESSION_OR_TIMEOUT);
		}

		ServerSideSessionInfo sinfo =
			(ServerSideSessionInfo)sessions.getSession(req.sessionId);

		if (sinfo.keyExchangeMode != Common.KEY_EXCHANGE_CIPHER_SERVER){
			main.logError(ServerLog.ERRLOG_LEVEL_NORMAL,
			"Bad server-cipher key exchange request");
			return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
					Util.encodeString("Invalid session id. Session failed."));
		}

		boolean oldClient = !req.hasEqualOrGreaterVersion((byte)2, (byte)2);
		byte[] encSessionKey = req.getEncryptedSessionKey();
		byte[] sessionKey = null;

		// this resonse will have protocol version the lesser of client and server
		AbstractResponse successResponse = new GenericResponse(req, AbstractMessage.RC_SUCCESS);
		
		// decrypt with server rsa private key
		try {
			sessionKey = Util.decrypt(privateKey, encSessionKey, successResponse.majorProtocolVersion, successResponse.minorProtocolVersion);
		} catch (Exception e) {
			return new ErrorResponse(req, AbstractMessage.RC_SESSION_FAILED,
					Util.encodeString("Can't decrypt client session key."));
		}

		int sessionKeyAlg = HdlSecurityProvider.ENCRYPT_ALG_DES;
		if(!oldClient) {
			// the caller is capable of encoding the encryption algorithm
			sessionKeyAlg = Encoder.readInt(sessionKey, 0);
			sessionKey = Util.substring(sessionKey, Encoder.INT_SIZE);
		}

		// replace the session key in the session info
		sinfo.setSessionKey(sessionKey);
		sinfo.setEncryptionAlgorithmCode(sessionKeyAlg);
		return successResponse;
	}

	//used for any request.
	private boolean validSession(AbstractRequest req)
	{
		if (req.sessionId == 0) return false;

		return getSession(req.sessionId)!=null;
	}

	// used to check for admin request:
	// if the session has been "challege response" authenticated, return true;
	// otherwise return false
	private boolean authenticatedSession(AbstractRequest req) {

	  // if we are allowing wide-open access, shortcut the authentication process
	  if(wideOpenAccessAllowed) return true;

		if (req.sessionId == 0) return false;

		ServerSideSessionInfo sssinfo = getSession(req.sessionId);
    return sssinfo!=null && !sssinfo.isSessionAnonymous() && sssinfo.clientAuthenticated;
	}

	//set the session given by session id as client authenticated.
	//please take note that all the req.authInfo is null in server side (this method)
	//even for administrative request.
	//this is because the decode message process can't re-produce the authInfo
	//on server side.
	//please see Encoder.java, all decodeMessage methods.
	//we can only compare the auth handle and auth index in challenge answer request
	//and in the session info.
	private void setSessionAuthenticated(AbstractRequest req, ChallengeAnswerRequest caReq, boolean authenticated){
		ServerSideSessionInfo ssinfo = null;
		if (req != null && req.sessionId > 0){
			ssinfo = getSession(req.sessionId);
		}
		if (ssinfo != null) {
			if (authenticated) {
				//we have to have the identityHandle and index not null

				if (ssinfo.identityKeyHandle == null || ssinfo.identityKeyIndex < 0) {
					//can not set the authenticated flag
					return;  //don't set the authenticated flag to an anonymous session
				} else {
					//if the session has identity handle and index set,
					//check if they match in the request
					if (!Util.equals(ssinfo.identityKeyHandle, caReq.userIdHandle) ||
							ssinfo.identityKeyIndex != caReq.userIdIndex) {
						return;  //don't set the authenticated flag to true if the identity doesn't match in the req
					}
				}
			}

			//set the authenticated flag
			ssinfo.clientAuthenticated = authenticated;
			if(DEBUG_AUTHENTICATION) System.err.println(new Date() + " session " + ssinfo.sessionId + " authenticated " + ssinfo.identityKeyIndex + ":" + Util.decodeString(ssinfo.identityKeyHandle));
		}
		return;
	}

	public ServerSideSessionInfo getSession(int sessionId){
		return sessions.getSession(sessionId);
	}


	/** Returns an object that can be synchronized in order to avoid performing
	 * conflicting operations on the given handle.  */
	private Object getWriteLock(byte hdl[]) {
		// quickly construct an index into the lockHash which is based on
		// the local part of the handle (case insensitive)
		int idx = 0;
		for(int i=hdl.length-1; i>=0; i--) {
			if(hdl[i]=='/') break;
			idx += Character.toLowerCase((char)hdl[i]);
		}
		idx = Math.abs(idx);
		return lockHash[idx%lockHash.length];
	}

	public final void shutdown() {
		keepRunning = false;
		if(replicationDaemon!=null) replicationDaemon.shutdown();
		if(sessions!=null) sessions.shutdown();
		if(txnQueue!=null) txnQueue.shutdown();
		if(storage!=null) storage.shutdown();
	}
	
	private boolean storageHaveNa(byte[] authHandle) throws HandleException {
	    if(storage.haveNA(authHandle)) return true;
	    if(Util.startsWithCI(authHandle,Common.NA_HANDLE_PREFIX)) return storage.haveNA(Util.getSuffixPart(authHandle));
	    if(Util.hasSlash(authHandle)) return false;
	    return storage.haveNA(Util.getZeroNAHandle(authHandle));
	}

}
