/**********************************************************************\
 © COPYRIGHT 2013 Corporation for National Research Initiatives (CNRI);
                        All rights reserved.
               
        The HANDLE.NET software is made available subject to the
         Handle System Public License, which may be obtained at
         http://hdl.handle.net/4263537/5030 or hdl:4263537/5030
\**********************************************************************/

package net.handle.server.replication;

import java.io.File;
import java.util.Collections;
import java.util.Iterator;
import java.util.NoSuchElementException;

import net.handle.hdllib.Encoder;
import net.handle.hdllib.HandleException;
import net.handle.server.Main;
import net.handle.server.ServerLog;

import com.sleepycat.je.Cursor;
import com.sleepycat.je.Database;
import com.sleepycat.je.DatabaseConfig;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.Environment;
import com.sleepycat.je.EnvironmentConfig;
import com.sleepycat.je.OperationStatus;
import com.sleepycat.je.PreloadConfig;

public class ReplicationDb {
    private static final String HANDLE_DB_NAME = "handlesDates";
    private static final String NA_DB_NAME = "nasDates";
    
    private Environment environment;
    private Database db;
    private Database naDB;
    private Object CREATE_OR_DELETE_LOCK = new Object();
    private Object HOME_OR_UNHOME_LOCK = new Object();

    private Main main;
    
    public ReplicationDb(File configDir, Main main) throws HandleException {
        this.main = main;
        File dbDir = new File(configDir,"replicationDb");
        if(!dbDir.exists()) dbDir.mkdirs();
        EnvironmentConfig envConfig = new EnvironmentConfig();
        envConfig.setTransactional(true);
        envConfig.setAllowCreate(true);
        envConfig.setLockTimeout(0);
        envConfig.setTxnWriteNoSync(false);
        envConfig.setSharedCache(true);
        try {
            environment = new Environment(dbDir, envConfig);
            db = openReplicationDb(HANDLE_DB_NAME);
            naDB = openReplicationDb(NA_DB_NAME);
        }
        catch(Exception e) {
            HandleException he = new HandleException(HandleException.CONFIGURATION_ERROR,"Unable to open replication database");
            he.initCause(e);
            throw he;
        }
    }
    
    public void deleteAll() throws Exception {
        if(db!=null) db = deleteReplicationDb(db,HANDLE_DB_NAME);
        if(naDB!=null) naDB = deleteReplicationDb(naDB,NA_DB_NAME);
    }
    
    private Database openReplicationDb(String dbName) throws DatabaseException {
        DatabaseConfig dbConfig = new DatabaseConfig();
        dbConfig.setTransactional(true);
        dbConfig.setAllowCreate(true);
        dbConfig.setSortedDuplicates(false);
        com.sleepycat.je.Transaction openTxn = environment.beginTransaction(null, null);
        Database db = environment.openDatabase(openTxn, dbName, dbConfig);
        openTxn.commit();
        db.preload(new PreloadConfig());
        return db;
    }
    
    private Database deleteReplicationDb(Database db,String dbName) throws DatabaseException {
        Database tmpDB = db;
        com.sleepycat.je.Transaction killTxn = environment.beginTransaction(null, null);
        try {
            db = null;
            tmpDB.close();
            tmpDB = null;
            environment.truncateDatabase(killTxn, dbName, false);
            killTxn.commitSync();
        } catch (DatabaseException t) {
            try { killTxn.abort(); } catch (Throwable t2) {}
            throw t;
        } finally {
            if(db==null) // try to re-open the database, if necessary
                db = openReplicationDb(dbName);
        }
        return db;
    }
    
    private class DBIterator implements Iterator {
        private Cursor cursor;
        private OperationStatus status;
        private DatabaseEntry key = new DatabaseEntry();
        private DatabaseEntry data = new DatabaseEntry();
        
        public DBIterator(Database db) {
            try {
                this.cursor = db.openCursor(null,null);
                status = cursor.getFirst(key,data,null);
            }
            catch(DatabaseException e) {
                System.err.println("Error in ReplicationDBIterator: " + e);
                status = null;
                close();
            }
        }
        
        public boolean hasNext() {
            if(status==null) return false;
            if(status==OperationStatus.SUCCESS) return true;
            close();
            status = null;
            return false;
        }
        public Object next() {
            if(!hasNext()) throw new NoSuchElementException();
            byte[] handle = key.getData();
            byte[] buf = new byte[16 + handle.length];
            Encoder.writeInt(buf,0,handle.length);
            System.arraycopy(handle,0,buf,4,handle.length);
            System.arraycopy(data.getData(),0,buf,4+handle.length,12);
            try {
                status = cursor.getNext(key,data,null);
            }
            catch(DatabaseException e) {
                System.err.println("Error in ReplicationDBIterator: " + e);
                status = null;
                close();
            }            
            return buf;
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
        public void finalize() {
            close();
        }
        public void close() {
            try {
                if (cursor != null) cursor.close();
            } catch (Throwable t) { }
        }
    }

    public Iterator/*<byte[]>*/ iterator(boolean isNA) {
        if(!isNA && db!=null) return new DBIterator(db);
        else if(isNA && naDB!=null) return new DBIterator(naDB);
        else return Collections.EMPTY_LIST.iterator();
    }
    
    public Object getLock(boolean isNA) {
        return isNA ? HOME_OR_UNHOME_LOCK : CREATE_OR_DELETE_LOCK;
    }
    
    public boolean isMoreRecentThanLastDate(byte[] handle, long date, int priority, boolean isNA) throws HandleException {
        Database db = isNA ? this.naDB : this.db;
        if(db!=null) {
            try {
                DatabaseEntry dbVal = new DatabaseEntry();
                if(db.get(null, new DatabaseEntry(handle), dbVal, null)==OperationStatus.SUCCESS) {
                    byte[] data = dbVal.getData();
                    long lastDate = Encoder.readLong(data,0);
                    int lastPriority = Encoder.readInt(data,8);
                    return (date > lastDate) || (date == lastDate && priority >= lastPriority);
                }
            }
            catch (DatabaseException e) {
                main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Database error in replication date handling: " + e);
                HandleException he = new HandleException(HandleException.INTERNAL_ERROR,"Database error in replication date handling");
                he.initCause(e);
                throw he;
            }
        }
        return true;
    }

    public long adjustAndSetLastDate(byte[] handle, long date, int priority, boolean isNA) throws HandleException {
        Database db = isNA ? this.naDB : this.db;
        synchronized(getLock(isNA)) {
            if(db!=null) {
                try {
                    DatabaseEntry dbVal = new DatabaseEntry();
                    if(db.get(null, new DatabaseEntry(handle), dbVal, null)==OperationStatus.SUCCESS) {
                        byte[] data = dbVal.getData();
                        long lastDate = Encoder.readLong(data,0);
                        int lastPriority = Encoder.readInt(data,8);
                        date = date <= lastDate ? lastDate + 1 : date;
                    }
                }
                catch (DatabaseException e) {
                    main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Database error in replication date handling: " + e);
                    HandleException he = new HandleException(HandleException.INTERNAL_ERROR,"Database error in replication date handling");
                    he.initCause(e);
                    throw he;
                }
            }
            setLastDate(handle,date,priority,isNA);
            return date;
        }
    }

    public void setLastDate(byte[] handle, long date, int priority, boolean isNA) throws HandleException {
        Database db = isNA ? this.naDB : this.db;
        if(db!=null) {
            synchronized(getLock(isNA)) {
                byte[] bytes = new byte[12];
                Encoder.writeLong(bytes,0,date);
                Encoder.writeInt(bytes,8,priority);
                try {
                    OperationStatus status = db.put(null, new DatabaseEntry(handle), new DatabaseEntry(bytes));
                    if(status!=OperationStatus.SUCCESS) {
                        main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Unknown status returned from db.put: "+status);
                        throw new HandleException(HandleException.INTERNAL_ERROR,"Database error in replication date handling");
                    }
                }
                catch (DatabaseException e) {
                    main.logError(ServerLog.ERRLOG_LEVEL_REALBAD,"Database error in replication date handling: " + e);
                    HandleException he = new HandleException(HandleException.INTERNAL_ERROR,"Database error in replication date handling");
                    he.initCause(e);
                    throw he;
                }
            }
        }
    }

    public void shutdown() {
        try {
            if(db!=null) db.close();
        }
        catch (Throwable e) {}
        try {
            if(naDB!=null) naDB.close();
        }
        catch (Throwable e) {}
        try {
            if(environment!=null) environment.close();
        }
        catch (Throwable e) {}        
    }

}
